from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField

from typing import Optional


class USINA(Register):
    """
    Registro que contém a usina modificada.
    """

    IDENTIFIER = " USINA"
    IDENTIFIER_DIGITS = 8
    LINE = Line([IntegerField(4, 9), LiteralField(20, 44)])

    @property
    def codigo(self) -> Optional[int]:
        """
        O principal conteúdo do registro (código da usina).

        :return: O código da usina
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo.setter
    def codigo(self, t: int):
        self.data[0] = t

    @property
    def nome(self) -> Optional[str]:
        """
        O nome da usina (opcional).

        :return: O nome da usina
        :rtype: Optional[str]
        """
        return self.data[1]

    @nome.setter
    def nome(self, t: str):
        self.data[1] = t


class VOLMIN(Register):
    """
    Registro que contém uma modificação de volume mínimo
    operativo para uma usina.
    """

    IDENTIFIER = " VOLMIN"
    IDENTIFIER_DIGITS = 8
    LINE = Line([LiteralField(30, 10)])
    # LINE = Line([FloatField(15, 10, 2), LiteralField(3, 26)])

    # @property
    # def volume(self) -> Optional[float]:
    #     """
    #     O novo valor de volume

    #     :return: O novo valor de volume
    #     :rtype: Optional[float]
    #     """
    #     return self.data[0]

    # @volume.setter
    # def volume(self, t: float):
    #     self.data[0] = t

    # @property
    # def unidade(self) -> Optional[str]:
    #     """
    #     A unidade do volume informado

    #     :return: A unidade do volume
    #     :rtype: Optional[str]
    #     """
    #     return self.data[1]

    # @unidade.setter
    # def unidade(self, t: str):
    #     self.data[1] = t


class VOLMAX(Register):
    """
    Registro que contém uma modificação de volume máximo
    operativo para uma usina.
    """

    IDENTIFIER = " VOLMAX"
    IDENTIFIER_DIGITS = 8
    LINE = Line([LiteralField(30, 10)])
    # LINE = Line([FloatField(15, 10, 2), LiteralField(3, 26)])

    # @property
    # def volume(self) -> Optional[float]:
    #     """
    #     O novo valor de volume

    #     :return: O novo valor de volume
    #     :rtype: Optional[float]
    #     """
    #     return self.data[0]

    # @volume.setter
    # def volume(self, t: float):
    #     self.data[0] = t

    # @property
    # def unidade(self) -> Optional[str]:
    #     """
    #     A unidade do volume informado

    #     :return: A unidade do volume
    #     :rtype: Optional[str]
    #     """
    #     return self.data[1]

    # @unidade.setter
    # def unidade(self, t: str):
    #     self.data[1] = t


class NUMCNJ(Register):
    """
    Registro que contém uma modificação de número de conjunto
    de máquinas.
    """

    IDENTIFIER = " NUMCNJ"
    IDENTIFIER_DIGITS = 8
    LINE = Line([IntegerField(3, 10)])

    @property
    def numero(self) -> int:
        """
        O novo valor do número de conjuntos

        :return: O novo número de conjuntos
        :rtype: Optional[int]
        """
        return self.data[0]

    @numero.setter
    def numero(self, t: int):
        self.data[0] = t


class NUMMAQ(Register):
    """
    Registro que contém uma modificação do número de máquinas em um
    conjunto de máquinas.
    """

    IDENTIFIER = " NUMMAQ"
    IDENTIFIER_DIGITS = 8
    LINE = Line([IntegerField(3, 10), IntegerField(3, 14)])

    @property
    def conjunto(self) -> Optional[int]:
        """
        O conjunto de máquinas que terá o número alterado

        :return: O índice do conjunto de máquinas
        :rtype: Optional[int]
        """
        return self.data[0]

    @conjunto.setter
    def conjunto(self, t: int):
        self.data[0] = t

    @property
    def numero_maquinas(self) -> Optional[int]:
        """
        O novo número de máquinas do conjunto

        :return: O número de máquinas do conjunto
        :rtype: Optional[int]
        """
        return self.data[1]

    @numero_maquinas.setter
    def numero_maquinas(self, t: int):
        self.data[1] = t


class VAZMIN(Register):
    """
    Registro que contém uma modificação de vazão mínima (m3/s).
    """

    IDENTIFIER = " VAZMIN "
    IDENTIFIER_DIGITS = 8
    LINE = Line([IntegerField(8, 10)])

    @property
    def vazao(self) -> Optional[int]:
        """
        O valor de vazão mínima

        :return: A nova vazão
        :rtype: Optional[int]
        """
        return self.data[0]

    @vazao.setter
    def vazao(self, t: int):
        self.data[0] = t


class CFUGA(Register):
    """
    Registro que contém uma modificação do nível do canal de fuga.
    """

    IDENTIFIER = " CFUGA"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [IntegerField(2, 10), IntegerField(4, 13), FloatField(7, 18, 3)]
    )

    @property
    def mes(self) -> int:
        """
        O mês de início da modificação

        :return: O mês de início da modificação
        :rtype: Optional[int]
        """
        return self.data[0]

    @mes.setter
    def mes(self, t: int):
        self.data[0] = t

    @property
    def ano(self) -> int:
        """
        O ano de início da modificação

        :return: O ano de início da modificação
        :rtype: Optional[int]
        """
        return self.data[1]

    @ano.setter
    def ano(self, t: int):
        self.data[1] = t

    @property
    def nivel(self) -> float:
        """
        O novo nivel do canal de fuga

        :return: O novo nível
        :rtype: Optional[int]
        """
        return self.data[2]

    @nivel.setter
    def nivel(self, t: float):
        self.data[2] = t


class CMONT(Register):
    """
    Registro que contém uma modificação do nível do canal de montante.
    """

    IDENTIFIER = " CMONT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [IntegerField(2, 10), IntegerField(4, 13), FloatField(7, 18, 3)]
    )

    @property
    def mes(self) -> Optional[int]:
        """
        O mês de início da modificação

        :return: O mês
        :rtype: Optional[int]
        """
        return self.data[0]

    @mes.setter
    def mes(self, t: int):
        self.data[0] = t

    @property
    def ano(self) -> Optional[int]:
        """
        O ano de início da modificação

        :return: O ano
        :rtype: Optional[int]
        """
        return self.data[1]

    @ano.setter
    def ano(self, t: int):
        self.data[1] = t

    @property
    def nivel(self) -> Optional[float]:
        """
        O novo nivel do canal de montante

        :return: O novo nível
        :rtype: Optional[float]
        """
        return self.data[2]

    @nivel.setter
    def nivel(self, t: float):
        self.data[2] = t


class VMAXT(Register):
    """
    Registro que contém uma modificação do volume máximo
    com data.
    """

    IDENTIFIER = " VMAXT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 3),
            LiteralField(3, 26),
        ]
    )

    @property
    def mes(self) -> Optional[int]:
        """
        O mês de início da modificação

        :return: O mês
        :rtype: Optional[int]
        """
        return self.data[0]

    @mes.setter
    def mes(self, t: int):
        self.data[0] = t

    @property
    def ano(self) -> Optional[int]:
        """
        O ano de início da modificação

        :return: O ano
        :rtype: Optional[int]
        """
        return self.data[1]

    @ano.setter
    def ano(self, t: int):
        self.data[1] = t

    @property
    def volume(self) -> Optional[float]:
        """
        O novo volume máximo

        :return: O novo volume
        :rtype: Optional[float]
        """
        return self.data[2]

    @volume.setter
    def volume(self, t: float):
        self.data[2] = t

    @property
    def unidade(self) -> Optional[str]:
        """
        A unidade de fornecimento do volume

        :return: A unidade
        :rtype: Optional[str]
        """
        return self.data[3]

    @unidade.setter
    def unidade(self, t: str):
        self.data[3] = t


class VMINT(Register):
    """
    Registro que contém uma modificação do volume mínimo
    com data.
    """

    IDENTIFIER = " VMINT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 3),
            LiteralField(3, 26),
        ]
    )

    @property
    def mes(self) -> Optional[int]:
        """
        O mês de início da modificação

        :return: O mês
        :rtype: Optional[int]
        """
        return self.data[0]

    @mes.setter
    def mes(self, t: int):
        self.data[0] = t

    @property
    def ano(self) -> Optional[int]:
        """
        O ano de início da modificação

        :return: O ano
        :rtype: Optional[int]
        """
        return self.data[1]

    @ano.setter
    def ano(self, t: int):
        self.data[1] = t

    @property
    def volume(self) -> Optional[float]:
        """
        O novo volume mínimo

        :return: O novo volume
        :rtype: Optional[float]
        """
        return self.data[2]

    @volume.setter
    def volume(self, t: float):
        self.data[2] = t

    @property
    def unidade(self) -> Optional[str]:
        """
        A unidade de fornecimento do volume

        :return: A unidade
        :rtype: Optional[str]
        """
        return self.data[3]

    @unidade.setter
    def unidade(self, t: str):
        self.data[3] = t


class VMINP(Register):
    """
    Registro que contém uma modificação do volume mínimo
    com data para adoção de penalidade.
    """

    IDENTIFIER = " VMINP"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 3),
            LiteralField(3, 26),
        ]
    )

    @property
    def mes(self) -> Optional[int]:
        """
        O mês de início da modificação

        :return: O mês
        :rtype: Optional[int]
        """
        return self.data[0]

    @mes.setter
    def mes(self, t: int):
        self.data[0] = t

    @property
    def ano(self) -> Optional[int]:
        """
        O ano de início da modificação

        :return: O ano
        :rtype: Optional[int]
        """
        return self.data[1]

    @ano.setter
    def ano(self, t: int):
        self.data[1] = t

    @property
    def volume(self) -> Optional[float]:
        """
        O novo volume mínimo a partir da data

        :return: O novo volume
        :rtype: Optional[float]
        """
        return self.data[2]

    @volume.setter
    def volume(self, t: float):
        self.data[2] = t

    @property
    def unidade(self) -> Optional[str]:
        """
        A unidade do volume fornecido

        :return: A unidade
        :rtype: Optional[str]
        """
        return self.data[3]

    @unidade.setter
    def unidade(self, t: str):
        self.data[3] = t


class VAZMINT(Register):
    """
    Registro que contém uma modificação da vazão mínima
    com data.
    """

    IDENTIFIER = " VAZMINT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 2),
        ]
    )

    @property
    def mes(self) -> Optional[int]:
        """
        O mês de início da modificação

        :return: O mês
        :rtype: Optional[int]
        """
        return self.data[0]

    @mes.setter
    def mes(self, t: int):
        self.data[0] = t

    @property
    def ano(self) -> Optional[int]:
        """
        O ano de início da modificação

        :return: O ano
        :rtype: Optional[int]
        """
        return self.data[1]

    @ano.setter
    def ano(self, t: int):
        self.data[1] = t

    @property
    def vazao(self) -> Optional[float]:
        """
        A nova vazão mínima a partir da data

        :return: A nova vazão
        :rtype: Optional[float]
        """
        return self.data[2]

    @vazao.setter
    def vazao(self, t: float):
        self.data[2] = t


class VAZMAXT(Register):
    """
    Registro que contém uma modificação da vazão máxima
    com data.
    """

    IDENTIFIER = " VAZMAXT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 2),
        ]
    )

    @property
    def mes(self) -> Optional[int]:
        """
        O mês de início da modificação

        :return: O mês
        :rtype: Optional[int]
        """
        return self.data[0]

    @mes.setter
    def mes(self, t: int):
        self.data[0] = t

    @property
    def ano(self) -> Optional[int]:
        """
        O ano de início da modificação

        :return: O ano
        :rtype: Optional[int]
        """
        return self.data[1]

    @ano.setter
    def ano(self, t: int):
        self.data[1] = t

    @property
    def vazao(self) -> Optional[float]:
        """
        A nova vazão máxima a partir da data

        :return: A nova vazão
        :rtype: Optional[float]
        """
        return self.data[2]

    @vazao.setter
    def vazao(self, t: float):
        self.data[2] = t


class TURBMAXT(Register):
    """
    Registro que contém uma modificação da turbinamento máximo
    com data.
    """

    IDENTIFIER = " TURBMAXT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 2),
        ]
    )

    @property
    def mes(self) -> Optional[int]:
        """
        O mês de início da modificação

        :return: O mês
        :rtype: Optional[int]
        """
        return self.data[0]

    @mes.setter
    def mes(self, t: int):
        self.data[0] = t

    @property
    def ano(self) -> Optional[int]:
        """
        O ano de início da modificação

        :return: O ano
        :rtype: Optional[int]
        """
        return self.data[1]

    @ano.setter
    def ano(self, t: int):
        self.data[1] = t

    @property
    def turbinamento(self) -> Optional[float]:
        """
        O novo turbinamento máximo a partir da data

        :return: O novo turbinamento
        :rtype: Optional[float]
        """
        return self.data[2]

    @turbinamento.setter
    def turbinamento(self, t: float):
        self.data[2] = t


class TURBMINT(Register):
    """
    Registro que contém uma modificação da turbinamento mínimo
    com data.
    """

    IDENTIFIER = " TURBMINT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 2),
        ]
    )

    @property
    def mes(self) -> Optional[int]:
        """
        O mês de início da modificação

        :return: O mês
        :rtype: Optional[int]
        """
        return self.data[0]

    @mes.setter
    def mes(self, t: int):
        self.data[0] = t

    @property
    def ano(self) -> Optional[int]:
        """
        O ano de início da modificação

        :return: O ano
        :rtype: Optional[int]
        """
        return self.data[1]

    @ano.setter
    def ano(self, t: int):
        self.data[1] = t

    @property
    def turbinamento(self) -> Optional[float]:
        """
        O novo turbinamento mínimo a partir da data

        :return: O novo turbinamento
        :rtype: Optional[float]
        """
        return self.data[2]

    @turbinamento.setter
    def turbinamento(self, t: float):
        self.data[2] = t



from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField

from typing import IO


class REE(Block):
    """
    Bloco com a informaçao do REE associado aos valores.
    """

    BEGIN_PATTERN = r"REE:"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line([LiteralField(12, 63)])

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, REE):
            return False
        bloco: REE = o
        if not all(
            [
                isinstance(self.data, str),
                isinstance(o.data, str),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):
        self.data = self.__linha.read(file.readline())[0]



MAX_SERIES_SINTETICAS = 2000
MAX_PATAMARES = 5
NUM_CENARIOS = 2000
MAX_ANOS_ESTUDO = 30
MAX_ANOS_HISTORICO = 120
MAX_CONFIGURACOES = 360
MAX_SUBMERCADOS = 15
MAX_ITERS = 100
MAX_FORWARDS = 300
MAX_CORTES = MAX_ITERS * MAX_FORWARDS
MAX_UHES = 330
MAX_CONJUNTOS_MAQUINAS = 5
MAX_UTES = 300
MAX_UEES = 999
MAX_RES = 100
MAX_REES = 30
MAX_LAG_ADTERM = 2
NUM_CONFIGS_DGER = 100
NUM_VARIAVEIS_MEDIAS = 52
ORDEM_MAX_PARP = 11
MESES = [
    "JANEIRO",
    "FEVEREIRO",
    "MARCO",
    "ABRIL",
    "MAIO",
    "JUNHO",
    "JULHO",
    "AGOSTO",
    "SETEMBRO",
    "OUTUBRO",
    "NOVEMBRO",
    "DEZEMBRO",
]
MESES_DF = [
    "Janeiro",
    "Fevereiro",
    "Março",
    "Abril",
    "Maio",
    "Junho",
    "Julho",
    "Agosto",
    "Setembro",
    "Outubro",
    "Novembro",
    "Dezembro",
]
MESES_ABREV = [
    "JAN",
    "FEV",
    "MAR",
    "ABR",
    "MAI",
    "JUN",
    "JUL",
    "AGO",
    "SET",
    "OUT",
    "NOV",
    "DEZ",
]
MAX_MESES_ESTUDO = MAX_ANOS_ESTUDO * len(MESES)



from inewave.config import (
    MAX_ANOS_ESTUDO,
    MAX_SUBMERCADOS,
    MESES_DF,
)

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class BlocoCargasAdicionais(Section):
    """
    Bloco com informações de cargas adicionais por mês/ano
    e por subsistema.
    """

    FIM_BLOCO = " 999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line(
            [IntegerField(3, 1), LiteralField(10, 6), LiteralField(12, 21)]
        )
        campo_ano: List[Field] = [LiteralField(4, 0)]
        campos_cargas: List[Field] = [
            FloatField(8, 6 + 8 * i, 0) for i in range(len(MESES_DF))
        ]
        self.__linha_cargas = Line(campo_ano + campos_cargas)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCargasAdicionais):
            return False
        bloco: BlocoCargasAdicionais = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=MESES_DF)
            df["Código Subsistema"] = codigo_subsis
            df["Nome Subsistema"] = nome_subsis
            df["Razão"] = razao
            df["Ano"] = anos
            df = df[
                ["Código Subsistema", "Nome Subsistema", "Razão", "Ano"]
                + MESES_DF
            ]
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        # Variáveis auxiliares
        codigo_subsis: List[int] = []
        nome_subsis: List[str] = []
        razao: List[str] = []
        anos: List[str] = []
        codigo_atual = 0
        nome_atual = ""
        razao_atual = ""
        tabela = np.zeros((MAX_ANOS_ESTUDO * MAX_SUBMERCADOS, len(MESES_DF)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 3:
                break
            if BlocoCargasAdicionais.FIM_BLOCO in linha:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            if len(linha.strip()) < 80:
                (
                    codigo_atual,
                    nome_atual,
                    razao_atual,
                ) = self.__linha_subsis.read(linha)
            else:
                dados = self.__linha_cargas.read(linha)
                tabela[i, :] = dados[1:]
                codigo_subsis.append(codigo_atual)
                nome_subsis.append(nome_atual)
                razao.append(razao_atual)
                anos.append(dados[0])
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        ultimo_codigo = 0
        ultimo_subsis = ""
        ultima_razao = ""
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do c_adic.dat não foram lidos com sucesso")

        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            if any(
                [
                    linha_lida["Código Subsistema"] != ultimo_codigo,
                    linha_lida["Nome Subsistema"] != ultimo_subsis,
                    linha_lida["Razão"] != ultima_razao,
                ]
            ):
                ultimo_codigo = linha_lida["Código Subsistema"]
                ultimo_subsis = linha_lida["Nome Subsistema"]
                ultima_razao = linha_lida["Razão"]
                file.write(
                    self.__linha_subsis.write(
                        linha_lida[
                            [
                                "Código Subsistema",
                                "Nome Subsistema",
                                "Razão",
                            ]
                        ].tolist()
                    )
                )
            linha_saida = linha_lida[["Ano"] + MESES_DF]
            valores_saida = []
            for valor, vazio in zip(linha_saida, linha_saida.isna()):
                valores_saida.append(None if vazio else valor)
            file.write(self.__linha_cargas.write(valores_saida))
        file.write(BlocoCargasAdicionais.FIM_BLOCO)



# Imports de módulos externos
from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.literalfield import LiteralField
from datetime import timedelta
import pandas as pd  # type: ignore
from typing import IO, List


class BlocoTemposEtapasTim(Block):
    """
    Bloco com as informações de tempo de execução
    do NEWAVE obtidas no arquivo `newave.tim`.
    """

    BEGIN_PATTERN = "Leitura de Dados:"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        # Cria a estrutura de uma linha da tabela
        etapa_field: List[Field] = [LiteralField(25, 0)]
        tempo_field: List[Field] = [LiteralField(14, 25)]
        self.__line = Line(etapa_field + tempo_field)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoTemposEtapasTim):
            return False
        bloco: BlocoTemposEtapasTim = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(data={"Etapa": etapas, "Tempo": tempos})
            return df

        # Variáveis auxiliares
        etapas: List[str] = []
        tempos: List[timedelta] = []

        # Leitura das etapas
        for _ in range(5):
            dados = self.__line.read(file.readline())
            etapas.append(dados[0].split(":")[0])
            h = int(dados[1].split("h")[0])
            min = int(dados[1].split("h")[1].split("min")[0])
            s = float(dados[1].split("min")[1].split("s")[0])
            ts = timedelta(hours=h, minutes=min, seconds=s)
            tempos.append(ts)

        self.data = converte_tabela_em_df()



from typing import Type, TypeVar, List, Optional, Union
from datetime import datetime

from cfinterface.components.register import Register
from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.eolicahistorico import (
    RegistroEolicaHistoricoVentoHorizonte,
    RegistroEolicaHistoricoVento,
)


class EolicaHistorico(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao histórico
    de ventos das usinas eólicas.
    """

    T = TypeVar("T")

    REGISTERS = [
        RegistroEolicaHistoricoVentoHorizonte,
        RegistroEolicaHistoricoVento,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="hist-ventos.csv"
    ) -> "EolicaHistorico":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="hist-ventos.csv"):
        self.write(diretorio, nome_arquivo)

    def __registros_por_tipo(self, registro: Type[T]) -> List[T]:
        """
        Obtém os registro de um tipo, se houver algum no arquivo.

        :param registro: Um tipo de registro para ser lido
        :type registro: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int

        """
        return [b for b in self.data.of_type(registro)]

    def __obtem_registros(self, tipo: Type[T]) -> List[T]:
        return self.__registros_por_tipo(tipo)

    def __obtem_registros_com_filtros(
        self, tipo_registro: Type[T], **kwargs
    ) -> Optional[Union[T, List[T]]]:
        def __atende(r) -> bool:
            condicoes: List[bool] = []
            for k, v in kwargs.items():
                if v is not None:
                    condicoes.append(getattr(r, k) == v)
            return all(condicoes)

        regs_filtro = [
            r for r in self.__obtem_registros(tipo_registro) if __atende(r)
        ]
        if len(regs_filtro) == 0:
            return None
        elif len(regs_filtro) == 1:
            return regs_filtro[0]
        else:
            return regs_filtro

    def cria_registro(self, anterior: Register, registro: Register):
        """
        Adiciona um registro ao arquivo após um outro registro previamente
        existente.

        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.add_after(anterior, registro)

    def deleta_registro(self, registro: Register):
        """
        Remove um registro existente no arquivo.

        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.remove(registro)

    def append_registro(self, registro: Register):
        """
        Adiciona um registro ao arquivo na última posição.
        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.append(registro)

    def preppend_registro(self, registro: Register):
        """
        Adiciona um registro ao arquivo na primeira posição.
        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.preppend(registro)

    def eolica_historico_vento_horizonte(
        self,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
    ) -> Optional[
        Union[
            RegistroEolicaHistoricoVentoHorizonte,
            List[RegistroEolicaHistoricoVentoHorizonte],
        ]
    ]:
        """
        Obtém um registro que contém o horizonte de um histórico.

        :param data_inicial: data de início do histórico
        :type data_inicial: datetime | None
        :param data_final: data de fim do histórico
        :type data_final: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaHistoricoVentoHorizonte` |
            list[:class:`RegistroEolicaHistoricoVentoHorizonte`] | None
        """
        return self.__obtem_registros_com_filtros(
            RegistroEolicaHistoricoVentoHorizonte,
            data_inicial=data_inicial,
            data_final=data_final,
        )

    def eolica_historico_vento(
        self,
        codigo_eolica: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        velocidade: Optional[float] = None,
        direcao: Optional[float] = None,
    ) -> Optional[
        Union[
            RegistroEolicaHistoricoVento,
            List[RegistroEolicaHistoricoVento],
        ]
    ]:
        """
        Obtém um registro que contém a o valor de vento para um
        intervalo do histórico.

        :param codigo_eolica: código da usina eólica
        :type codigo_eolica: int | None
        :param data_inicial: data de início do registro histórico
        :type data_inicial: datetime | None
        :param data_final: data de fim do registro histórico
        :type data_final: datetime | None
        :param velocidade: velocidade do vento
        :type velocidade: float | None
        :param direcao: direção do vento
        :type direcao: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaHistoricoVento` |
            list[:class:`RegistroEolicaHistoricoVento`] | None
        """
        return self.__obtem_registros_com_filtros(
            RegistroEolicaHistoricoVento,
            codigo_eolica=codigo_eolica,
            data_inicial=data_inicial,
            data_final=data_final,
            velocidade=velocidade,
            direcao=direcao,
        )



from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type, Optional

import pandas as pd  # type: ignore

from inewave.newave.modelos.ree import (
    BlocoReesSubmercados,
    BlocoFicticiasIndividualizado,
)


class REE(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às configurações
    dos REEs.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = [
        BlocoReesSubmercados,
        BlocoFicticiasIndividualizado,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="ree.dat") -> "REE":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="ree.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def rees(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os REES e os submercados

        - Número (`int`)
        - Nome (`str`)
        - Submercado (`int`)
        - Mês Fim Individualizado (`int`)
        - Ano Fim Individualizado (`int`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoReesSubmercados, 0)
        if b is not None:
            return b.data
        return None

    @rees.setter
    def rees(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoReesSubmercados, 0)
        if b is not None:
            b.data = df

    @property
    def remocao_ficticias(self) -> Optional[int]:
        """
        Opção de remover usinas fictícias nos períodos individualizados.

        :return: O valor do campo.
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoFicticiasIndividualizado, 0)
        if b is not None:
            return b.data[1]
        return None

    @remocao_ficticias.setter
    def remocao_ficticias(self, d: int):
        b = self.__bloco_por_tipo(BlocoFicticiasIndividualizado, 0)
        if b is not None:
            b.data[1] = d



from inewave.config import MAX_REES

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class BlocoPenalidades(Section):
    """
    Bloco com informações das penalidades por violação de restrições
    e outras ações.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                LiteralField(6, 1),
                FloatField(8, 14, 2),
                FloatField(8, 24, 2),
                IntegerField(3, 36),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoPenalidades):
            return False
        bloco: BlocoPenalidades = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            cols = [
                "Penalidade 1",
                "Penalidade 2",
                "Subsistema",
            ]
            df = pd.DataFrame(tabela, columns=cols)
            df["Chave"] = chaves
            df = df[
                [
                    "Chave",
                    "Penalidade 1",
                    "Penalidade 2",
                    "Subsistema",
                ]
            ]
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        i = 0
        chaves: List[str] = []
        # TODO - remover o 5 mágico abaixo
        tabela = np.zeros((MAX_REES * 5, 3))
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            else:
                dados = self.__linha.read(linha)
                tabela[i, :] = dados[1:]
                chaves.append(dados[0])
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do penalid.dat não foram lidos com sucesso"
            )

        for _, lin in self.data.iterrows():
            linha_escrita = lin.tolist()
            for i, v in enumerate(linha_escrita[1:], start=1):
                linha_escrita[i] = None if np.isnan(v) else v
            if linha_escrita[3] is not None:
                linha_escrita[3] = int(linha_escrita[3])
            file.write(self.__linha.write(linha_escrita))



# Rotinas de testes associadas ao arquivo dger.dat do NEWAVE
from inewave.newave.dger import DGer

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.dger import MockDger


def test_atributos_nao_encontrados_dger():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.nome_caso == ""
        assert d.tipo_execucao is None
        assert d.duracao_periodo is None
        assert d.num_anos_estudo is None
        assert d.mes_inicio_pre_estudo is None
        assert d.mes_inicio_estudo is None
        assert d.ano_inicio_estudo is None
        assert d.num_anos_pre_estudo is None
        assert d.num_anos_pos_estudo is None
        assert d.num_anos_pos_sim_final is None
        assert d.imprime_dados is None
        assert d.imprime_mercados is None
        assert d.imprime_energias is None
        assert d.imprime_modelo_estocastico is None
        assert d.imprime_subsistema is None
        assert d.num_max_iteracoes is None
        assert d.num_forwards is None
        assert d.num_aberturas is None
        assert d.num_series_sinteticas is None
        assert d.ordem_maxima_parp is None
        assert d.ano_inicial_historico is None
        assert d.tamanho_registro_arquivo_historico is None
        assert d.calcula_volume_inicial is None
        assert d.volume_inicial_subsistema == [
            None,
            None,
            None,
            None,
            None,
        ]
        assert d.tolerancia is None
        assert d.taxa_de_desconto is None
        assert d.tipo_simulacao_final is None
        assert d.agregacao_simulacao_final is None
        assert d.impressao_operacao is None
        assert d.impressao_convergencia is None
        assert d.intervalo_para_gravar is None
        assert d.num_minimo_iteracoes is None
        assert d.racionamento_preventivo is None
        assert d.num_anos_manutencao_utes is None
        assert d.considera_tendencia_hidrologica_calculo_politica is None
        assert d.considera_tendencia_hidrologica_sim_final is None
        assert d.restricao_itaipu is None
        assert d.bid is None
        assert d.perdas_rede_transmissao is None
        assert d.el_nino is None
        assert d.enso is None
        assert d.duracao_por_patamar is None
        assert d.outros_usos_da_agua is None
        assert d.correcao_desvio is None
        assert d.curva_aversao is None
        assert d.tipo_geracao_enas is None
        assert d.primeira_profundidade_risco_deficit is None
        assert d.segunda_profundidade_risco_deficit is None
        assert d.iteracao_para_simulacao_final is None
        assert d.agrupamento_livre is None
        assert d.equalizacao_penal_intercambio is None
        assert d.representacao_submotorizacao is None
        assert d.ordenacao_automatica is None
        assert d.considera_carga_adicional is None
        assert d.delta_zsup is None
        assert d.delta_zinf is None
        assert d.deltas_consecutivos is None
        assert d.despacho_antecipado_gnl is None
        assert d.modif_automatica_adterm is None
        assert d.considera_ghmin is None
        assert d.simulacao_final_com_data is None
        assert d.utiliza_gerenciamento_pls is None
        assert d.comunicacao_dois_niveis is None
        assert d.armazenamento_local_arquivos_temporarios is None
        assert d.alocacao_memoria_ena is None
        assert d.alocacao_memoria_cortes is None
        assert d.sar is None
        assert d.cvar is None
        assert d.considera_zsup_min_convergencia is None
        assert d.desconsidera_vazao_minima is None
        assert d.restricoes_eletricas is None
        assert d.selecao_de_cortes_backward is None
        assert d.selecao_de_cortes_forward is None
        assert d.janela_de_cortes is None
        assert d.considera_reamostragem_cenarios is None
        assert d.tipo_reamostragem_cenarios is None
        assert d.passo_reamostragem_cenarios is None
        assert d.converge_no_zero is None
        assert d.consulta_fcf is None
        assert d.impressao_ena is None
        assert d.impressao_cortes_ativos_sim_final is None
        assert d.representacao_agregacao is None
        assert d.matriz_correlacao_espacial is None
        assert d.desconsidera_convergencia_estatistica is None
        assert d.momento_reamostragem is None
        assert d.mantem_arquivos_energias is None
        assert d.inicio_teste_convergencia is None
        assert d.sazonaliza_vmint is None
        assert d.sazonaliza_vmaxt is None
        assert d.sazonaliza_vminp is None
        assert d.sazonaliza_cfuga_cmont is None
        assert d.restricoes_emissao_gee is None
        assert d.consideracao_media_anual_afluencias is None
        assert d.reducao_automatica_ordem is None
        assert d.restricoes_fornecimento_gas is None
        assert d.memoria_calculo_cortes is None
        assert d.considera_geracao_eolica is None
        assert d.penalidade_corte_geracao_eolica is None
        assert d.compensacao_correlacao_cruzada is None
        assert d.restricao_defluencia is None
        assert d.restricao_turbinamento is None
        assert d.aproveitamento_bases_backward is None
        assert d.impressao_estados_geracao_cortes is None
        assert d.semente_forward is None
        assert d.semente_backward is None
        assert d.restricao_lpp_turbinamento_maximo_ree is None
        assert d.restricao_lpp_defluencia_maxima_ree is None
        assert d.restricao_lpp_turbinamento_maximo_uhe is None
        assert d.restricao_lpp_defluencia_maxima_uhe is None


def test_atributos_encontrados_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.nome_caso is not None
        assert d.tipo_execucao is not None
        assert d.duracao_periodo is not None
        assert d.num_anos_estudo is not None
        assert d.mes_inicio_pre_estudo is not None
        assert d.mes_inicio_estudo is not None
        assert d.ano_inicio_estudo is not None
        assert d.num_anos_pre_estudo is not None
        assert d.num_anos_pos_estudo is not None
        assert d.num_anos_pos_sim_final is not None
        assert d.imprime_dados is not None
        assert d.imprime_mercados is not None
        assert d.imprime_energias is not None
        assert d.imprime_modelo_estocastico is not None
        assert d.imprime_subsistema is not None
        assert d.num_max_iteracoes is not None
        assert d.num_forwards is not None
        assert d.num_aberturas is not None
        assert d.num_series_sinteticas is not None
        assert d.ordem_maxima_parp is not None
        assert d.ano_inicial_historico is not None
        assert d.tamanho_registro_arquivo_historico is not None
        assert d.calcula_volume_inicial is not None
        assert d.volume_inicial_subsistema != [
            None,
            None,
            None,
            None,
            None,
        ]
        assert d.tolerancia is not None
        assert d.taxa_de_desconto is not None
        assert d.tipo_simulacao_final is not None
        assert d.agregacao_simulacao_final is not None
        assert d.impressao_operacao is not None
        assert d.impressao_convergencia is not None
        assert d.intervalo_para_gravar is not None
        assert d.num_minimo_iteracoes is not None
        assert d.racionamento_preventivo is not None
        assert d.num_anos_manutencao_utes is not None
        assert d.considera_tendencia_hidrologica_calculo_politica is not None
        assert d.considera_tendencia_hidrologica_sim_final is not None
        assert d.restricao_itaipu is not None
        assert d.bid is not None
        assert d.perdas_rede_transmissao is not None
        assert d.el_nino is not None
        assert d.enso is not None
        assert d.duracao_por_patamar is not None
        assert d.outros_usos_da_agua is not None
        assert d.correcao_desvio is not None
        assert d.curva_aversao is not None
        assert d.tipo_geracao_enas is not None
        assert d.primeira_profundidade_risco_deficit is not None
        assert d.segunda_profundidade_risco_deficit is not None
        assert d.iteracao_para_simulacao_final is not None
        assert d.agrupamento_livre is not None
        assert d.equalizacao_penal_intercambio is not None
        assert d.representacao_submotorizacao is not None
        assert d.ordenacao_automatica is not None
        assert d.considera_carga_adicional is not None
        assert d.delta_zsup is not None
        assert d.delta_zinf is not None
        assert d.deltas_consecutivos is not None
        assert d.despacho_antecipado_gnl is not None
        assert d.modif_automatica_adterm is not None
        assert d.considera_ghmin is not None
        assert d.simulacao_final_com_data is not None
        assert d.utiliza_gerenciamento_pls is not None
        assert d.comunicacao_dois_niveis is not None
        assert d.armazenamento_local_arquivos_temporarios is not None
        assert d.alocacao_memoria_ena is not None
        assert d.alocacao_memoria_cortes is not None
        assert d.sar is not None
        assert d.cvar is not None
        assert d.considera_zsup_min_convergencia is not None
        assert d.desconsidera_vazao_minima is not None
        assert d.restricoes_eletricas is not None
        assert d.selecao_de_cortes_backward is not None
        assert d.selecao_de_cortes_forward is not None
        assert d.janela_de_cortes is not None
        assert d.considera_reamostragem_cenarios is not None
        assert d.tipo_reamostragem_cenarios is not None
        assert d.passo_reamostragem_cenarios is not None
        assert d.converge_no_zero is not None
        assert d.consulta_fcf is not None
        assert d.impressao_ena is not None
        assert d.impressao_cortes_ativos_sim_final is not None
        assert d.representacao_agregacao is not None
        assert d.matriz_correlacao_espacial is not None
        assert d.desconsidera_convergencia_estatistica is not None
        assert d.momento_reamostragem is not None
        assert d.mantem_arquivos_energias is not None
        assert d.inicio_teste_convergencia is not None
        assert d.sazonaliza_vmint is not None
        assert d.sazonaliza_vmaxt is not None
        assert d.sazonaliza_vminp is not None
        assert d.sazonaliza_cfuga_cmont is not None
        assert d.restricoes_emissao_gee is not None
        assert d.consideracao_media_anual_afluencias is not None
        assert d.reducao_automatica_ordem is not None
        assert d.restricoes_fornecimento_gas is not None
        assert d.memoria_calculo_cortes is not None
        assert d.considera_geracao_eolica is not None
        assert d.penalidade_corte_geracao_eolica is not None
        assert d.compensacao_correlacao_cruzada is not None
        assert d.restricao_defluencia is not None
        assert d.restricao_turbinamento is not None
        assert d.aproveitamento_bases_backward is not None
        assert d.impressao_estados_geracao_cortes is not None
        assert d.semente_forward is not None
        assert d.semente_backward is not None
        assert d.restricao_lpp_turbinamento_maximo_ree is not None
        assert d.restricao_lpp_defluencia_maxima_ree is not None
        assert d.restricao_lpp_turbinamento_maximo_uhe is not None
        assert d.restricao_lpp_defluencia_maxima_uhe is not None


def test_nome_caso_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.nome_caso == (
            "PMO JANEIRO - 2021  22/12/2020 12:43:55  Niveis"
            + " para 26/12 NW Versao 27.5_CPAMP"
        )
        novo_valor = "Teste"
        d.nome_caso = novo_valor
        assert d.nome_caso == novo_valor


def test_tipo_execucao_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.tipo_execucao == 1
        novo_valor = 0
        d.tipo_execucao = novo_valor
        assert d.tipo_execucao == novo_valor


def test_duracao_periodo_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.duracao_periodo == 1
        novo_valor = 0
        d.duracao_periodo = novo_valor
        assert d.duracao_periodo == novo_valor


def test_num_anos_estudo_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.num_anos_estudo == 5
        novo_valor = 0
        d.num_anos_estudo = novo_valor
        assert d.num_anos_estudo == novo_valor


def test_mes_inicio_pre_estudo_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.mes_inicio_pre_estudo == 1
        novo_valor = 0
        d.mes_inicio_pre_estudo = novo_valor
        assert d.mes_inicio_pre_estudo == novo_valor


def test_mes_inicio_estudo_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.mes_inicio_estudo == 1
        novo_valor = 0
        d.mes_inicio_estudo = novo_valor
        assert d.mes_inicio_estudo == novo_valor


def test_ano_inicio_estudo_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.ano_inicio_estudo == 2021
        novo_valor = 0
        d.ano_inicio_estudo = novo_valor
        assert d.ano_inicio_estudo == novo_valor


def test_num_anos_pre_estudo_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.num_anos_pre_estudo == 0
        novo_valor = 5
        d.num_anos_pre_estudo = novo_valor
        assert d.num_anos_pre_estudo == novo_valor


def test_num_anos_pos_estudo_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.num_anos_pos_estudo == 5
        novo_valor = 0
        d.num_anos_pos_estudo = novo_valor
        assert d.num_anos_pos_estudo == novo_valor


def test_num_anos_pos_sim_final_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.num_anos_pos_sim_final == 0
        novo_valor = 1
        d.num_anos_pos_sim_final = novo_valor
        assert d.num_anos_pos_sim_final == novo_valor


def test_imprime_dados_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.imprime_dados == 1
        novo_valor = 0
        d.imprime_dados = novo_valor
        assert d.imprime_dados == novo_valor


def test_imprime_mercados_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.imprime_mercados == 1
        novo_valor = 0
        d.imprime_mercados = novo_valor
        assert d.imprime_mercados == novo_valor


def test_imprime_energias_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.imprime_energias == 1
        novo_valor = 0
        d.imprime_energias = novo_valor
        assert d.imprime_energias == novo_valor


def test_imprime_modelo_estocastico_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.imprime_modelo_estocastico == 1
        novo_valor = 0
        d.imprime_modelo_estocastico = novo_valor
        assert d.imprime_modelo_estocastico == novo_valor


def test_imprime_subsistema_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.imprime_subsistema == 1
        novo_valor = 0
        d.imprime_subsistema = novo_valor
        assert d.imprime_subsistema == novo_valor


def test_num_max_iteracoes_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.num_max_iteracoes == 45
        novo_valor = 0
        d.num_max_iteracoes = novo_valor
        assert d.num_max_iteracoes == novo_valor


def test_num_forwards_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.num_forwards == 200
        novo_valor = 0
        d.num_forwards = novo_valor
        assert d.num_forwards == novo_valor


def test_num_aberturas_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.num_aberturas == 20
        novo_valor = 0
        d.num_aberturas = novo_valor
        assert d.num_aberturas == novo_valor


def test_num_series_sinteticas_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.num_series_sinteticas == 2000
        novo_valor = 0
        d.num_series_sinteticas = novo_valor
        assert d.num_series_sinteticas == novo_valor


def test_ordem_maxima_parp_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.ordem_maxima_parp == 6
        novo_valor = 0
        d.ordem_maxima_parp = novo_valor
        assert d.ordem_maxima_parp == novo_valor


def test_ano_inicial_historico_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.ano_inicial_historico == 1931
        novo_valor = 0
        d.ano_inicial_historico = novo_valor
        assert d.ano_inicial_historico == novo_valor


def test_tamanho_registro_arquivo_historico_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.tamanho_registro_arquivo_historico == 0
        novo_valor = 1
        d.tamanho_registro_arquivo_historico = novo_valor
        assert d.tamanho_registro_arquivo_historico == novo_valor


def test_calcula_volume_inicial_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.calcula_volume_inicial == 1
        novo_valor = 0
        d.calcula_volume_inicial = novo_valor
        assert d.calcula_volume_inicial == novo_valor


def test_volume_inicial_subsistema_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.volume_inicial_subsistema == [0.0, 0.0, 0.0, 0.0, 0.0]
        novo_valor = [1.0, 1.0, 1.0, 1.0, 1.0]
        d.volume_inicial_subsistema = novo_valor
        assert d.volume_inicial_subsistema == novo_valor


def test_tolerancia_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.tolerancia == 95.0
        novo_valor = 0.0
        d.tolerancia = novo_valor
        assert d.tolerancia == novo_valor


def test_taxa_de_desconto_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.taxa_de_desconto == 12.0
        novo_valor = 0.0
        d.taxa_de_desconto = novo_valor
        assert d.taxa_de_desconto == novo_valor


def test_tipo_simulacao_final_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.tipo_simulacao_final == 1
        novo_valor = 0
        d.tipo_simulacao_final = novo_valor
        assert d.tipo_simulacao_final == novo_valor


def test_agregacao_simulacao_final_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.agregacao_simulacao_final == 1
        novo_valor = 0
        d.agregacao_simulacao_final = novo_valor
        assert d.agregacao_simulacao_final == novo_valor


def test_impressao_operacao_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.impressao_operacao == 1
        novo_valor = 0
        d.impressao_operacao = novo_valor
        assert d.impressao_operacao == novo_valor


def test_impressao_convergencia_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.impressao_convergencia == 1
        novo_valor = 0
        d.impressao_convergencia = novo_valor
        assert d.impressao_convergencia == novo_valor


def test_intervalo_para_gravar_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.intervalo_para_gravar == 1
        novo_valor = 0
        d.intervalo_para_gravar = novo_valor
        assert d.intervalo_para_gravar == novo_valor


def test_num_minimo_iteracoes_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.num_minimo_iteracoes == 30
        novo_valor = 0
        d.num_minimo_iteracoes = novo_valor
        assert d.num_minimo_iteracoes == novo_valor


def test_racionamento_preventivo_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.racionamento_preventivo == 0
        novo_valor = 1
        d.racionamento_preventivo = novo_valor
        assert d.racionamento_preventivo == novo_valor


def test_num_anos_manutencao_utes_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.num_anos_manutencao_utes == 1
        novo_valor = 0
        d.num_anos_manutencao_utes = novo_valor
        assert d.num_anos_manutencao_utes == novo_valor


def test_tendencia_hidrologica_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.considera_tendencia_hidrologica_calculo_politica == 2
        assert d.considera_tendencia_hidrologica_sim_final == 2
        novo_valor = 0
        d.considera_tendencia_hidrologica_calculo_politica = novo_valor
        assert d.considera_tendencia_hidrologica_calculo_politica == novo_valor
        d.considera_tendencia_hidrologica_sim_final = novo_valor
        assert d.considera_tendencia_hidrologica_sim_final == novo_valor


def test_restricao_itaipu_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.restricao_itaipu == 0
        novo_valor = 1
        d.restricao_itaipu = novo_valor
        assert d.restricao_itaipu == novo_valor


def test_bid_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.bid == 0
        novo_valor = 1
        d.bid = novo_valor
        assert d.bid == novo_valor


def test_perdas_rede_transmissao_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.perdas_rede_transmissao == 0
        novo_valor = 1
        d.perdas_rede_transmissao = novo_valor
        assert d.perdas_rede_transmissao == novo_valor


def test_el_nino_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.el_nino == 0
        novo_valor = 1
        d.el_nino = novo_valor
        assert d.el_nino == novo_valor


def test_enso_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.enso == 0
        novo_valor = 1
        d.enso = novo_valor
        assert d.enso == novo_valor


def test_duracao_por_patamar_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.duracao_por_patamar == 1
        novo_valor = 0
        d.duracao_por_patamar = novo_valor
        assert d.duracao_por_patamar == novo_valor


def test_outros_usos_da_agua_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.outros_usos_da_agua == 1
        novo_valor = 0
        d.outros_usos_da_agua = novo_valor
        assert d.outros_usos_da_agua == novo_valor


def test_correcao_desvio_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.correcao_desvio == 1
        novo_valor = 0
        d.correcao_desvio = novo_valor
        assert d.correcao_desvio == novo_valor


def test_curva_aversao_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.curva_aversao == 1
        novo_valor = 0
        d.curva_aversao = novo_valor
        assert d.curva_aversao == novo_valor


def test_tipo_geracao_enas_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.tipo_geracao_enas == 0
        novo_valor = 1
        d.tipo_geracao_enas = novo_valor
        assert d.tipo_geracao_enas == novo_valor


def test_risco_deficit_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.primeira_profundidade_risco_deficit == 1.0
        assert d.segunda_profundidade_risco_deficit == 2.5
        novo_valor = 0.0
        d.primeira_profundidade_risco_deficit = novo_valor
        assert d.primeira_profundidade_risco_deficit == novo_valor
        d.segunda_profundidade_risco_deficit = novo_valor
        assert d.segunda_profundidade_risco_deficit == novo_valor


def test_iteracao_para_simulacao_final_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.iteracao_para_simulacao_final == 0
        novo_valor = 1
        d.iteracao_para_simulacao_final = novo_valor
        assert d.iteracao_para_simulacao_final == novo_valor


def test_agrupamento_livre_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.agrupamento_livre == 1
        novo_valor = 0
        d.agrupamento_livre = novo_valor
        assert d.agrupamento_livre == novo_valor


def test_equalizacao_penal_intercambio_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.equalizacao_penal_intercambio == 1
        novo_valor = 0
        d.equalizacao_penal_intercambio = novo_valor
        assert d.equalizacao_penal_intercambio == novo_valor


def test_representacao_submotorizacao_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.representacao_submotorizacao == 2
        novo_valor = 0
        d.representacao_submotorizacao = novo_valor
        assert d.representacao_submotorizacao == novo_valor


def test_ordenacao_automatica_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.ordenacao_automatica == 0
        novo_valor = 1
        d.ordenacao_automatica = novo_valor
        assert d.ordenacao_automatica == novo_valor


def test_considera_carga_adicional_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.considera_carga_adicional == 1
        novo_valor = 0
        d.considera_carga_adicional = novo_valor
        assert d.considera_carga_adicional == novo_valor


def test_delta_zsup_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.delta_zsup == 10
        novo_valor = 0.0
        d.delta_zsup = novo_valor
        assert d.delta_zsup == novo_valor


def test_delta_zinf_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.delta_zinf == 0.2
        novo_valor = 0.0
        d.delta_zinf = novo_valor
        assert d.delta_zinf == novo_valor


def test_deltas_consecutivos_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.deltas_consecutivos == 3
        novo_valor = 0
        d.deltas_consecutivos = novo_valor
        assert d.deltas_consecutivos == novo_valor


def test_despacho_antecipado_gnl_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.despacho_antecipado_gnl == 1
        novo_valor = 0
        d.despacho_antecipado_gnl = novo_valor
        assert d.despacho_antecipado_gnl == novo_valor


def test_modif_automatica_adterm_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.modif_automatica_adterm == 1
        novo_valor = 0
        d.modif_automatica_adterm = novo_valor
        assert d.modif_automatica_adterm == novo_valor


def test_considera_ghmin_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.considera_ghmin == 1
        novo_valor = 0
        d.considera_ghmin = novo_valor
        assert d.considera_ghmin == novo_valor


def test_simulacao_final_com_data_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.simulacao_final_com_data == 0
        novo_valor = 1
        d.simulacao_final_com_data = novo_valor
        assert d.simulacao_final_com_data == novo_valor


def test_gerenciamento_pls_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.utiliza_gerenciamento_pls == 0
        assert d.comunicacao_dois_niveis == 0
        assert d.armazenamento_local_arquivos_temporarios == 0
        assert d.alocacao_memoria_ena == 0
        assert d.alocacao_memoria_cortes == 0
        novo_valor = 1
        d.utiliza_gerenciamento_pls = novo_valor
        assert d.utiliza_gerenciamento_pls == novo_valor
        d.comunicacao_dois_niveis = novo_valor
        assert d.comunicacao_dois_niveis == novo_valor
        d.armazenamento_local_arquivos_temporarios = novo_valor
        assert d.armazenamento_local_arquivos_temporarios == novo_valor
        d.alocacao_memoria_ena = novo_valor
        assert d.alocacao_memoria_ena == novo_valor
        d.alocacao_memoria_cortes = novo_valor
        assert d.alocacao_memoria_cortes == novo_valor


def test_sar_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.sar == 0
        novo_valor = 1
        d.sar = novo_valor
        assert d.sar == novo_valor


def test_cvar_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.cvar == 1
        novo_valor = 0
        d.cvar = novo_valor
        assert d.cvar == novo_valor


def test_considera_zsup_min_convergencia_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.considera_zsup_min_convergencia == 0
        novo_valor = 0
        d.considera_zsup_min_convergencia = novo_valor
        assert d.considera_zsup_min_convergencia == novo_valor


def test_desconsidera_vazao_minima_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.desconsidera_vazao_minima == 0
        novo_valor = 1
        d.desconsidera_vazao_minima = novo_valor
        assert d.desconsidera_vazao_minima == novo_valor


def test_restricoes_eletricas_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.restricoes_eletricas == 1
        novo_valor = 0
        d.restricoes_eletricas = novo_valor
        assert d.restricoes_eletricas == novo_valor


def test_selecao_de_cortes_backward_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.selecao_de_cortes_backward == 1
        novo_valor = 0
        d.selecao_de_cortes_backward = novo_valor
        assert d.selecao_de_cortes_backward == novo_valor


def test_selecao_de_cortes_forward_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.selecao_de_cortes_forward == 1
        novo_valor = 0
        d.selecao_de_cortes_forward = novo_valor
        assert d.selecao_de_cortes_forward == novo_valor


def test_janela_de_cortes_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.janela_de_cortes == 0
        novo_valor = 1
        d.janela_de_cortes = novo_valor
        assert d.janela_de_cortes == novo_valor


def test_reamostragem_cenarios_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.considera_reamostragem_cenarios == 1
        assert d.tipo_reamostragem_cenarios == 1
        assert d.passo_reamostragem_cenarios == 1
        novo_valor = 0
        d.considera_reamostragem_cenarios = novo_valor
        assert d.considera_reamostragem_cenarios == novo_valor
        d.tipo_reamostragem_cenarios = novo_valor
        assert d.tipo_reamostragem_cenarios == novo_valor
        d.passo_reamostragem_cenarios = novo_valor
        assert d.passo_reamostragem_cenarios == novo_valor


def test_converge_no_zero_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.converge_no_zero == 0
        novo_valor = 1
        d.converge_no_zero = novo_valor
        assert d.converge_no_zero == novo_valor


def test_consulta_fcf_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.consulta_fcf == 0
        novo_valor = 1
        d.consulta_fcf = novo_valor
        assert d.consulta_fcf == novo_valor


def test_impressao_ena_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.impressao_ena == 1
        novo_valor = 0
        d.impressao_ena = novo_valor
        assert d.impressao_ena == novo_valor


def test_impressao_cortes_ativos_sim_final_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.impressao_cortes_ativos_sim_final == 0
        novo_valor = 1
        d.impressao_cortes_ativos_sim_final = novo_valor
        assert d.impressao_cortes_ativos_sim_final == novo_valor


def test_representacao_agregacao_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.representacao_agregacao == 1
        novo_valor = 0
        d.representacao_agregacao = novo_valor
        assert d.representacao_agregacao == novo_valor


def test_matriz_correlacao_espacial_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.matriz_correlacao_espacial == 1
        novo_valor = 0
        d.matriz_correlacao_espacial = novo_valor
        assert d.matriz_correlacao_espacial == novo_valor


def test_desconsidera_convergencia_estatistica_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.desconsidera_convergencia_estatistica == 1
        novo_valor = 0
        d.desconsidera_convergencia_estatistica = novo_valor
        assert d.desconsidera_convergencia_estatistica == novo_valor


def test_momento_reamostragem_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.momento_reamostragem == 1
        novo_valor = 0
        d.momento_reamostragem = novo_valor
        assert d.momento_reamostragem == novo_valor


def test_mantem_arquivos_energias_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.mantem_arquivos_energias == 0
        novo_valor = 1
        d.mantem_arquivos_energias = novo_valor
        assert d.mantem_arquivos_energias == novo_valor


def test_inicio_teste_convergencia_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.inicio_teste_convergencia == 1
        novo_valor = 0
        d.inicio_teste_convergencia = novo_valor
        assert d.inicio_teste_convergencia == novo_valor


def test_sazonaliza_vmint_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.sazonaliza_vmint == 0
        novo_valor = 1
        d.sazonaliza_vmint = novo_valor
        assert d.sazonaliza_vmint == novo_valor


def test_sazonaliza_vmaxt_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.sazonaliza_vmaxt == 0
        novo_valor = 1
        d.sazonaliza_vmaxt = novo_valor
        assert d.sazonaliza_vmaxt == novo_valor


def test_sazonaliza_vminp_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.sazonaliza_vminp == 0
        novo_valor = 1
        d.sazonaliza_vminp = novo_valor
        assert d.sazonaliza_vminp == novo_valor


def test_sazonaliza_cfuga_cmont_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.sazonaliza_cfuga_cmont == 0
        novo_valor = 1
        d.sazonaliza_cfuga_cmont = novo_valor
        assert d.sazonaliza_cfuga_cmont == novo_valor


def test_restricoes_emissao_gee_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.restricoes_emissao_gee == 0
        novo_valor = 1
        d.restricoes_emissao_gee = novo_valor
        assert d.restricoes_emissao_gee == novo_valor


def test_afluencia_anual_parp_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.consideracao_media_anual_afluencias == 3
        assert d.reducao_automatica_ordem == 0
        novo_valor = 1
        d.consideracao_media_anual_afluencias = novo_valor
        assert d.consideracao_media_anual_afluencias == novo_valor
        d.reducao_automatica_ordem = novo_valor
        assert d.reducao_automatica_ordem == novo_valor


def test_restricoes_fornecimento_gas_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.restricoes_fornecimento_gas == 0
        novo_valor = 1
        d.restricoes_fornecimento_gas = novo_valor
        assert d.restricoes_fornecimento_gas == novo_valor


def test_memoria_calculo_cortes_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.memoria_calculo_cortes == 0
        novo_valor = 1
        d.memoria_calculo_cortes = novo_valor
        assert d.memoria_calculo_cortes == novo_valor


def test_considera_geracao_eolica_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.considera_geracao_eolica == 1
        novo_valor = 0
        d.considera_geracao_eolica = novo_valor
        assert d.considera_geracao_eolica == novo_valor


def test_penalidade_corte_geracao_eolica_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.penalidade_corte_geracao_eolica == 0.0063
        novo_valor = 1.0
        d.penalidade_corte_geracao_eolica = novo_valor
        assert d.penalidade_corte_geracao_eolica == novo_valor


def test_compensacao_correlacao_cruzada_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.compensacao_correlacao_cruzada == 1
        novo_valor = 0
        d.compensacao_correlacao_cruzada = novo_valor
        assert d.compensacao_correlacao_cruzada == novo_valor


def test_restricao_turbinamento_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.restricao_turbinamento == 1
        novo_valor = 0
        d.restricao_turbinamento = novo_valor
        assert d.restricao_turbinamento == novo_valor


def test_restricao_defluencia_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.restricao_defluencia == 1
        novo_valor = 0
        d.restricao_defluencia = novo_valor
        assert d.restricao_defluencia == novo_valor


def test_aproveitamento_base_pls_backward_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.aproveitamento_bases_backward == 1
        novo_valor = 0
        d.aproveitamento_bases_backward = novo_valor
        assert d.aproveitamento_bases_backward == novo_valor


def test_impressao_estados_geracao_cortes_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.impressao_estados_geracao_cortes == 1
        novo_valor = 0
        d.impressao_estados_geracao_cortes = novo_valor
        assert d.impressao_estados_geracao_cortes == novo_valor


def test_semente_forward_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.semente_forward == 0
        novo_valor = 1000
        d.semente_forward = novo_valor
        assert d.semente_forward == novo_valor


def test_semente_backward_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.semente_backward == 0
        novo_valor = 1000
        d.semente_backward = novo_valor
        assert d.semente_backward == novo_valor


def test_restricao_lpp_turbinamento_maximo_ree_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.restricao_lpp_turbinamento_maximo_ree == 1
        novo_valor = 0
        d.restricao_lpp_turbinamento_maximo_ree = novo_valor
        assert d.restricao_lpp_turbinamento_maximo_ree == novo_valor


def test_restricao_lpp_defluencia_maxima_ree_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.restricao_lpp_defluencia_maxima_ree == 1
        novo_valor = 0
        d.restricao_lpp_defluencia_maxima_ree = novo_valor
        assert d.restricao_lpp_defluencia_maxima_ree == novo_valor


def test_restricao_lpp_turbinamento_maximo_uhe_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.restricao_lpp_turbinamento_maximo_uhe == 1
        novo_valor = 0
        d.restricao_lpp_turbinamento_maximo_uhe = novo_valor
        assert d.restricao_lpp_turbinamento_maximo_uhe == novo_valor


def test_restricao_lpp_defluencia_maxima_uhe_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d = DGer.le_arquivo("")
        assert d.restricao_lpp_defluencia_maxima_uhe == 1
        novo_valor = 0
        d.restricao_lpp_defluencia_maxima_uhe = novo_valor
        assert d.restricao_lpp_defluencia_maxima_uhe == novo_valor


def test_eq_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d1 = DGer.le_arquivo("")
        d2 = DGer.le_arquivo("")
        assert d1 == d2


def test_neq_dger():
    m: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m):
        d1 = DGer.le_arquivo("")
        d2 = DGer.le_arquivo("")
        d2.nome_caso = "Teste"
        assert d1 != d2


def test_leitura_escrita_dger():
    m_leitura: MagicMock = mock_open(read_data="".join(MockDger))
    with patch("builtins.open", m_leitura):
        d1 = DGer.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        d1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        d2 = DGer.le_arquivo("")
        assert d1 == d2



# Rotinas de testes associadas ao arquivo ree.dat do NEWAVE
from inewave.newave.ree import REE


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.ree import (
    MockREE,
)


def test_atributos_encontrados_ree():
    m: MagicMock = mock_open(read_data="".join(MockREE))
    with patch("builtins.open", m):
        ad = REE.le_arquivo("")
        assert ad.rees is not None
        assert ad.remocao_ficticias is not None


def test_atributos_nao_encontrados_ree():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = REE.le_arquivo("")
        assert ad.rees is None
        assert ad.remocao_ficticias is None


def test_eq_ree():
    m: MagicMock = mock_open(read_data="".join(MockREE))
    with patch("builtins.open", m):
        cf1 = REE.le_arquivo("")
        cf2 = REE.le_arquivo("")
        assert cf1 == cf2


def test_neq_ree():
    m: MagicMock = mock_open(read_data="".join(MockREE))
    with patch("builtins.open", m):
        cf1 = REE.le_arquivo("")
        cf2 = REE.le_arquivo("")
        cf2.rees.loc[0, 0] = 0
        assert cf1 != cf2


def test_leitura_escrita_ree():
    m_leitura: MagicMock = mock_open(read_data="".join(MockREE))
    with patch("builtins.open", m_leitura):
        cf1 = REE.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = REE.le_arquivo("")
        assert cf1 == cf2



from inewave.nwlistcf.modelos.nwlistcf import CortesPeriodoNwlistcf

from cfinterface.files.blockfile import BlockFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Nwlistcf(BlockFile):
    """
    Armazena os dados dos cortes construídos pelo NEWAVE existentes
    no arquivo `nwlistcf.rel` do NWLISTCF.

    Esta classe armazena os cortes da FCF de cada uma das variáveis,
    para cada registro e REE dentro do registro.

    """

    T = TypeVar("T")

    BLOCKS = [CortesPeriodoNwlistcf]

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__cortes_periodos = None

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="nwlistcf.rel"
    ) -> "Nwlistcf":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="nwlistcf.rel"):
        self.write(diretorio, nome_arquivo)

    def __monta_tabela_cortes(self) -> pd.DataFrame:
        df = None
        for b in self.data.of_type(CortesPeriodoNwlistcf):
            dados = b.data
            if dados is None:
                continue
            elif df is None:
                df = b.data
            else:
                df = pd.concat([df, b.data], ignore_index=True)
        return df

    @property
    def cortes(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os cortes da FCF.

        - PERIODO (`int`)
        - IREG (`int`)
        - REE (`int`)
        - RHS (`int`)
        - PIV (`int`)
        - PIH 1 (`float`)
        - ...
        - PIH 6 (`float`)
        - PIGTAD(P1L1) (`float`)
        - ...
        - PIGTAD(P3L2) (`float`)
        - PIMX_SAR (`float`)
        - PIMX_CURVA (`float`)

        :return: A tabela de cortes como um DataFrame
        :rtype: pd.DataFrame | None
        """
        if self.__cortes_periodos is None:
            self.__cortes_periodos = self.__monta_tabela_cortes()
        return self.__cortes_periodos



# Rotinas de testes associadas ao arquivo clast.dat do NEWAVE
from inewave.newave.modelos.clast import BlocoUTEClasT
from inewave.newave.modelos.clast import BlocoModificacaoUTEClasT

from inewave.newave import ClasT


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.clast import MockBlocoUTEClasT
from tests.mocks.arquivos.clast import MockBlocoModificacaoClasT
from tests.mocks.arquivos.clast import MockClasT


def test_bloco_ute_clast():
    m: MagicMock = mock_open(read_data="".join(MockBlocoUTEClasT))
    b = BlocoUTEClasT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, -1] == 0.0


def test_atributos_encontrados_ute_clast():
    m: MagicMock = mock_open(read_data="".join(MockBlocoUTEClasT))
    with patch("builtins.open", m):
        ct = ClasT.le_arquivo("")
        assert ct.usinas is not None


def test_atributos_nao_encontrados_ute_clast():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ct = ClasT.le_arquivo("")
        assert ct.usinas is None


def test_bloco_modificacao_clast():
    m: MagicMock = mock_open(read_data="".join(MockBlocoModificacaoClasT))
    b = BlocoModificacaoUTEClasT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 211
    assert b.data.iloc[-1, -1] == "PORTO ITAQUI"


def test_atributos_encontrados_modificacao_clast():
    m: MagicMock = mock_open(read_data="".join(MockClasT))
    with patch("builtins.open", m):
        ct = ClasT.le_arquivo("")
        assert ct.modificacoes is not None


def test_atributos_nao_encontrados_modificacao_clast():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ct = ClasT.le_arquivo("")
        assert ct.modificacoes is None


def test_eq_clast():
    m: MagicMock = mock_open(read_data="".join(MockClasT))
    with patch("builtins.open", m):
        ct1 = ClasT.le_arquivo("")
        ct2 = ClasT.le_arquivo("")
        assert ct1 == ct2


def test_neq_cadic():
    m: MagicMock = mock_open(read_data="".join(MockClasT))
    with patch("builtins.open", m):
        ct1 = ClasT.le_arquivo("")
        ct2 = ClasT.le_arquivo("")
        ct2.usinas.iloc[0, 0] = -1
        assert ct1 != ct2


def test_leitura_escrita_clast():
    m_leitura: MagicMock = mock_open(read_data="".join(MockClasT))
    with patch("builtins.open", m_leitura):
        ct1 = ClasT.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        ct1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        ct2 = ClasT.le_arquivo("")
        assert ct1 == ct2



# Rotinas de testes associadas ao arquivo eolica-config.csv do NEWAVE
from datetime import datetime
from inewave.newave.modelos.eolicaconfiguracao import (
    RegistroEolicaConfiguracao,
)

from inewave.newave.eolicaconfiguracao import EolicaConfiguracao

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.eolicaconfig import (
    MockRegistroEolicaConfiguracaoPeriodo,
    MockEolicaConfig,
)


def test_registro_eolica_configuracao_periodo_eolicaconfig():

    m: MagicMock = mock_open(
        read_data="".join(MockRegistroEolicaConfiguracaoPeriodo)
    )
    r = RegistroEolicaConfiguracao()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [
        1,
        datetime(2021, 1, 1),
        datetime(2030, 12, 1),
        "centralizado",
    ]
    assert r.codigo_eolica == 1
    r.codigo_eolica = 5
    assert r.data_inicial_estado_operacao == datetime(2021, 1, 1)
    r.data_inicial_estado_operacao = datetime(2021, 2, 1)
    assert r.data_final_estado_operacao == datetime(2030, 12, 1)
    r.data_final_estado_operacao = datetime(2030, 11, 1)
    assert r.estado_operacao == "centralizado"
    r.estado_operacao = "fixo"


def test_atributos_encontrados_eolicaconfig():
    m: MagicMock = mock_open(read_data="".join(MockEolicaConfig))
    with patch("builtins.open", m):
        e = EolicaConfiguracao.le_arquivo("")
        assert len(e.eolica_configuracao()) > 0


def test_eq_eolicaconfig():
    m: MagicMock = mock_open(read_data="".join(MockEolicaConfig))
    with patch("builtins.open", m):
        cf1 = EolicaConfiguracao.le_arquivo("")
        cf2 = EolicaConfiguracao.le_arquivo("")
        assert cf1 == cf2


def test_neq_eolicaconfig():
    m: MagicMock = mock_open(read_data="".join(MockEolicaConfig))
    with patch("builtins.open", m):
        cf1 = EolicaConfiguracao.le_arquivo("")
        cf2 = EolicaConfiguracao.le_arquivo("")
        cf2.deleta_registro(cf1.eolica_configuracao()[0])
        assert cf1 != cf2


def test_leitura_escrita_eolicaconfig():
    m_leitura: MagicMock = mock_open(read_data="".join(MockEolicaConfig))
    with patch("builtins.open", m_leitura):
        cf1 = EolicaConfiguracao.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = EolicaConfiguracao.le_arquivo("")
        assert cf1 == cf2



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore

from inewave.config import MAX_ANOS_ESTUDO, MAX_UHES, MESES_DF


class BlocoDsvUHE(Section):
    """
    Bloco de informações do desvio de água por
    usina no arquivo do NEWAVE `dsvagua.dat`.
    """

    FIM_BLOCO = "9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campos_iniciais: List[Field] = [IntegerField(4, 0), IntegerField(3, 6)]
        campos_desvios: List[Field] = [
            FloatField(6, 10 + 7 * i, 2) for i in range(len(MESES_DF))
        ]
        campos_finais: List[Field] = [
            IntegerField(4, 94),
            LiteralField(33, 103),
        ]
        self.__linha_uhe = Line(
            campos_iniciais + campos_desvios + campos_finais
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoDsvUHE):
            return False
        bloco: BlocoDsvUHE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            cols = ["Ano", "Usina"] + MESES_DF + ["Flag"]
            df = pd.DataFrame(
                tabela,
                columns=cols,
            )
            df["Comentário"] = comentarios
            df = df.astype({"Ano": "int64", "Usina": "int64", "Flag": "int64"})
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        i = 0
        tabela = np.zeros((MAX_UHES * MAX_ANOS_ESTUDO, len(MESES_DF) + 3))
        comentarios: List[str] = []
        while True:
            linha = file.readline()
            # Confere se terminaram as usinas
            if len(linha) < 3 or BlocoDsvUHE.FIM_BLOCO in linha:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            dados = self.__linha_uhe.read(linha)
            tabela[i, :] = dados[: len(MESES_DF) + 3]
            comentarios.append(dados[-1])
            i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do confhd.dat não foram lidos com sucesso")

        for _, lin in self.data.iterrows():
            file.write(self.__linha_uhe.write(lin.tolist()))
        file.write(BlocoDsvUHE.FIM_BLOCO + "\n")



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore


class BlocoConfUHE(Section):
    """
    Bloco de informações das usinas cadastradas
    no arquivo do NEWAVE `confhd.dat`.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_uhe = Line(
            [
                IntegerField(4, 1),
                LiteralField(12, 6),
                IntegerField(4, 19),
                IntegerField(4, 25),
                IntegerField(4, 30),
                FloatField(6, 35, 2),
                LiteralField(4, 42),
                IntegerField(4, 49),
                IntegerField(4, 58),
                IntegerField(4, 67),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoConfUHE):
            return False
        bloco: BlocoConfUHE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def extrai_coluna_de_listas(listas: List[list], coluna: int) -> list:
            return [lista[coluna] for lista in listas]

        def transforma_uhes_em_tabela() -> pd.DataFrame:
            # Converte as informações de cada linha em colunas
            col_num = extrai_coluna_de_listas(dados_uhes, 0)
            col_nome = extrai_coluna_de_listas(dados_uhes, 1)
            col_posto = extrai_coluna_de_listas(dados_uhes, 2)
            col_jus = extrai_coluna_de_listas(dados_uhes, 3)
            col_ree = extrai_coluna_de_listas(dados_uhes, 4)
            col_vinic = extrai_coluna_de_listas(dados_uhes, 5)
            col_exis = extrai_coluna_de_listas(dados_uhes, 6)
            col_modif = extrai_coluna_de_listas(dados_uhes, 7)
            col_inic_hist = extrai_coluna_de_listas(dados_uhes, 8)
            col_fim_hist = extrai_coluna_de_listas(dados_uhes, 9)
            dados = {
                "Número": col_num,
                "Nome": col_nome,
                "Posto": col_posto,
                "Jusante": col_jus,
                "REE": col_ree,
                "Volume Inicial": col_vinic,
                "Usina Existente": col_exis,
                "Modificada": col_modif,
                "Início do Histórico": col_inic_hist,
                "Fim do Histórico": col_fim_hist,
            }
            return pd.DataFrame(data=dados)

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        # Para cada usina, lê e processa as informações
        dados_uhes: List[list] = []
        while True:
            linha = file.readline()
            # Confere se terminaram as usinas
            if len(linha) < 3:
                # Converte para df e salva na variável
                if len(dados_uhes) > 0:
                    self.data = transforma_uhes_em_tabela()
                break
            dados_uhe = self.__linha_uhe.read(linha)
            dados_uhes.append(dados_uhe)

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do confhd.dat não foram lidos com sucesso")

        for _, lin in self.data.iterrows():
            file.write(self.__linha_uhe.write(lin.tolist()))



from inewave.config import MESES_DF

from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField

from inewave.nwlistop.modelos.blocos.valoresserie import ValoresSerie


class EafsAnos(ValoresSerie):
    """
    Bloco com as informações das tabelas de energias afluente bruta
    por série e por mês/ano de estudo.
    """

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [  # type: ignore
            IntegerField(4, 2),
        ]
        + [FloatField(8, 7 + 9 * i, 1) for i in range(len(MESES_DF) + 1)]  # type: ignore
    )



"""
inewave
=======

inewave é um módulo em Python para manipulação dos arquivos de entrada
e saída do NEWAVE.
"""

__version__ = "0.0.75"

from . import newave  # noqa
from . import nwlistcf  # noqa
from . import nwlistop  # noqa



from inewave.nwlistop.ctermsin import CtermSIN

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.ctermsin import MockCtermSIN


def test_atributos_encontrados_ctermsin():
    m: MagicMock = mock_open(read_data="".join(MockCtermSIN))
    with patch("builtins.open", m):
        n = CtermSIN.le_arquivo("")
        assert n.valores is not None
        assert n.valores.iloc[0, 0] == 2021
        assert n.valores.iloc[-1, -1] == 1230.35


def test_atributos_nao_encontrados_ctermsin():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        n = CtermSIN.le_arquivo("")
        assert n.valores is None


def test_eq_ctermsin():
    m: MagicMock = mock_open(read_data="".join(MockCtermSIN))
    with patch("builtins.open", m):
        n1 = CtermSIN.le_arquivo("")
        n2 = CtermSIN.le_arquivo("")
        assert n1 == n2


# Não deve ter teste de diferença, visto que o atributo é
# implementado como Lazy Property.



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore

from inewave.config import MAX_UHES, MESES_DF


class BlocoVazPast(Section):
    """
    Bloco de informações de vazões passadas
    por usina, existentes no arquivo `vazpast.dat`
    do NEWAVE.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [IntegerField(4, 1), LiteralField(12, 6)]
            + [FloatField(10, 18 + 10 * i, 2) for i in range(len(MESES_DF))]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoVazPast):
            return False
        bloco: BlocoVazPast = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            cols = ["Índice"] + MESES_DF
            df = pd.DataFrame(
                tabela,
                columns=cols,
            )
            df["Usina"] = usinas
            df = df.astype({"Índice": "int64"})
            df = df[["Índice", "Usina"] + MESES_DF]
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        tabela = np.zeros((MAX_UHES, len(MESES_DF) + 1))
        usinas: List[str] = []
        while True:
            linha = file.readline()
            # Confere se terminaram as usinas
            if len(linha) < 3:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            tabela[i, 0] = dados[0]
            usinas.append(dados[1])
            tabela[i, 1:] = dados[2:]
            i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do vazpast.dat não foram lidos com sucesso"
            )

        for _, lin in self.data.iterrows():
            file.write(self.__linha.write(lin.tolist()))



from inewave.newave.modelos.clast import (
    BlocoUTEClasT,
    BlocoModificacaoUTEClasT,
)

from cfinterface.files.sectionfile import SectionFile
from typing import Type, TypeVar, Optional
import pandas as pd  # type: ignore


class ClasT(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às classes de
    usinas térmicas.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoUTEClasT, BlocoModificacaoUTEClasT]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="clast.dat") -> "ClasT":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="clast.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def usinas(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as usinas e seus custos.

        - Número (`int`)
        - Nome (`str`)
        - Tipo Combustível (`str`)
        - Custo [1-5] (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoUTEClasT, 0)
        if b is not None:
            return b.data
        return None

    @usinas.setter
    def usinas(self, valor: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoUTEClasT, 0)
        if b is not None:
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def modificacoes(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as modificações de custos das usinas
        organizadas por usina.

        - Número (`int`)
        - Custo (`float`)
        - Mês Início (`int`)
        - Ano Início (`int`)
        - Mês Fim (`int`)
        - Ano Fim (`int`)
        - Nome (`str`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoModificacaoUTEClasT, 0)
        if b is not None:
            return b.data
        return None

    @modificacoes.setter
    def modificacoes(self, valor: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoModificacaoUTEClasT, 0)
        if b is not None:
            b.data = valor
        else:
            raise ValueError("Campo não lido")



from inewave.newave.modelos.cadic import BlocoCargasAdicionais

from inewave.newave import CAdic


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.cadic import MockBlocoCargasAdicionais


def test_bloco_ute_cadic():

    m: MagicMock = mock_open(read_data="".join(MockBlocoCargasAdicionais))
    b = BlocoCargasAdicionais()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 12
    assert b.data.shape[1] == 16
    assert b.data.iloc[0, 2] == "CONS.ITAIPU"
    assert b.data.iloc[-1, -1] == 2246


def test_atributos_encontrados_cadic():
    m: MagicMock = mock_open(read_data="".join(MockBlocoCargasAdicionais))
    with patch("builtins.open", m):
        ad = CAdic.le_arquivo("")
        assert ad.cargas is not None


def test_atributos_nao_encontrados_cadic():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = CAdic.le_arquivo("")
        assert ad.cargas is None


def test_eq_cadic():
    m: MagicMock = mock_open(read_data="".join(MockBlocoCargasAdicionais))
    with patch("builtins.open", m):
        ad1 = CAdic.le_arquivo("")
        ad2 = CAdic.le_arquivo("")
        assert ad1 == ad2


def test_neq_cadic():
    m: MagicMock = mock_open(read_data="".join(MockBlocoCargasAdicionais))
    with patch("builtins.open", m):
        ad1 = CAdic.le_arquivo("")
        ad2 = CAdic.le_arquivo("")
        ad2.cargas.iloc[0, 0] = -1
        assert ad1 != ad2


def test_leitura_escrita_cadic():
    m_leitura: MagicMock = mock_open(
        read_data="".join(MockBlocoCargasAdicionais)
    )
    with patch("builtins.open", m_leitura):
        ad1 = CAdic.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        ad1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        ad2 = CAdic.le_arquivo("")
        assert ad1 == ad2



from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List


class RegistroUHEHidr(Register):
    """
    Registro com os dados associados a uma UHE, existentes no
    arquivo HIDR.
    """

    LINE = Line(
        [
            LiteralField(size=12, starting_position=0),  # Nome
            IntegerField(size=4, starting_position=12),  # Posto
            IntegerField(size=8, starting_position=16),  # Posto BDH
            IntegerField(size=4, starting_position=24),  # Subsistema
            IntegerField(size=4, starting_position=28),  # Empresa
            IntegerField(size=4, starting_position=32),  # Jusante
            IntegerField(size=4, starting_position=36),  # Desvio
            FloatField(size=4, starting_position=40),  # Volume Mínimo
            FloatField(size=4, starting_position=44),  # Volume Máximo
            FloatField(size=4, starting_position=48),  # Volume Vertedouro
            FloatField(size=4, starting_position=52),  # Volume Desvio
            FloatField(size=4, starting_position=56),  # Cota Mínima
            FloatField(size=4, starting_position=60),  # Cota Máxima
            FloatField(size=4, starting_position=64),  # Pol VC 0
            FloatField(size=4, starting_position=68),  # Pol VC 1
            FloatField(size=4, starting_position=72),  # Pol VC 2
            FloatField(size=4, starting_position=76),  # Pol VC 3
            FloatField(size=4, starting_position=80),  # Pol VC 4
            FloatField(size=4, starting_position=84),  # Pol CA 0
            FloatField(size=4, starting_position=88),  # Pol CA 1
            FloatField(size=4, starting_position=92),  # Pol CA 2
            FloatField(size=4, starting_position=96),  # Pol CA 3
            FloatField(size=4, starting_position=100),  # Pol CA 4
            IntegerField(size=4, starting_position=104),  # Evap JAN
            IntegerField(size=4, starting_position=108),  # Evap FEV
            IntegerField(size=4, starting_position=112),  # Evap MAR
            IntegerField(size=4, starting_position=116),  # Evap ABR
            IntegerField(size=4, starting_position=120),  # Evap MAI
            IntegerField(size=4, starting_position=124),  # Evap JUN
            IntegerField(size=4, starting_position=128),  # Evap JUL
            IntegerField(size=4, starting_position=132),  # Evap AGO
            IntegerField(size=4, starting_position=136),  # Evap SET
            IntegerField(size=4, starting_position=140),  # Evap OUT
            IntegerField(size=4, starting_position=144),  # Evap NOV
            IntegerField(size=4, starting_position=148),  # Evap DEZ
            IntegerField(size=4, starting_position=152),  # Num. Conj. Máquinas
            IntegerField(
                size=4, starting_position=156
            ),  # Num. Máquinas Conj. 1
            IntegerField(
                size=4, starting_position=160
            ),  # Num. Máquinas Conj. 2
            IntegerField(
                size=4, starting_position=164
            ),  # Num. Máquinas Conj. 3
            IntegerField(
                size=4, starting_position=168
            ),  # Num. Máquinas Conj. 4
            IntegerField(
                size=4, starting_position=172
            ),  # Num. Máquinas Conj. 5
            FloatField(size=4, starting_position=176),  # Potência Conj. 1
            FloatField(size=4, starting_position=180),  # Potência Conj. 2
            FloatField(size=4, starting_position=184),  # Potência Conj. 3
            FloatField(size=4, starting_position=188),  # Potência Conj. 4
            FloatField(size=4, starting_position=192),  # Potência Conj. 5
            LiteralField(size=300, starting_position=196),  # Campos ignorados
            FloatField(size=4, starting_position=496),  # H Nominal 1
            FloatField(size=4, starting_position=500),  # H Nominal 2
            FloatField(size=4, starting_position=504),  # H Nominal 3
            FloatField(size=4, starting_position=508),  # H Nominal 4
            FloatField(size=4, starting_position=512),  # H Nominal 5
            IntegerField(size=4, starting_position=516),  # Q Nominal 1
            IntegerField(size=4, starting_position=520),  # Q Nominal 2
            IntegerField(size=4, starting_position=524),  # Q Nominal 3
            IntegerField(size=4, starting_position=528),  # Q Nominal 4
            IntegerField(size=4, starting_position=532),  # Q Nominal 5
            FloatField(size=4, starting_position=536),  # Produtibilidade
            FloatField(size=4, starting_position=540),  # Perdas
            IntegerField(size=4, starting_position=544),  # Núm. Pol. Jus.
            FloatField(size=4, starting_position=548),  # Pol. Jus. 1 - 0
            FloatField(size=4, starting_position=552),  # Pol. Jus. 1 - 1
            FloatField(size=4, starting_position=556),  # Pol. Jus. 1 - 2
            FloatField(size=4, starting_position=560),  # Pol. Jus. 1 - 3
            FloatField(size=4, starting_position=564),  # Pol. Jus. 1 - 4
            FloatField(size=4, starting_position=568),  # Pol. Jus. 2 - 0
            FloatField(size=4, starting_position=572),  # Pol. Jus. 2 - 1
            FloatField(size=4, starting_position=576),  # Pol. Jus. 2 - 2
            FloatField(size=4, starting_position=580),  # Pol. Jus. 2 - 3
            FloatField(size=4, starting_position=584),  # Pol. Jus. 2 - 4
            FloatField(size=4, starting_position=588),  # Pol. Jus. 3 - 0
            FloatField(size=4, starting_position=592),  # Pol. Jus. 3 - 1
            FloatField(size=4, starting_position=596),  # Pol. Jus. 3 - 2
            FloatField(size=4, starting_position=600),  # Pol. Jus. 3 - 3
            FloatField(size=4, starting_position=604),  # Pol. Jus. 3 - 4
            FloatField(size=4, starting_position=608),  # Pol. Jus. 4 - 0
            FloatField(size=4, starting_position=612),  # Pol. Jus. 4 - 1
            FloatField(size=4, starting_position=616),  # Pol. Jus. 4 - 2
            FloatField(size=4, starting_position=620),  # Pol. Jus. 4 - 3
            FloatField(size=4, starting_position=624),  # Pol. Jus. 4 - 4
            FloatField(size=4, starting_position=628),  # Pol. Jus. 5 - 0
            FloatField(size=4, starting_position=632),  # Pol. Jus. 5 - 1
            FloatField(size=4, starting_position=636),  # Pol. Jus. 5 - 2
            FloatField(size=4, starting_position=640),  # Pol. Jus. 5 - 3
            FloatField(size=4, starting_position=644),  # Pol. Jus. 5 - 4
            FloatField(size=4, starting_position=648),  # Pol. Jus. 6 - 0
            FloatField(size=4, starting_position=652),  # Pol. Jus. 6 - 1
            FloatField(size=4, starting_position=656),  # Pol. Jus. 6 - 2
            FloatField(size=4, starting_position=660),  # Pol. Jus. 6 - 3
            FloatField(size=4, starting_position=664),  # Pol. Jus. 6 - 4
            FloatField(size=4, starting_position=668),  # Pol. Jus. REF - 1
            FloatField(size=4, starting_position=672),  # Pol. Jus. REF - 2
            FloatField(size=4, starting_position=676),  # Pol. Jus. REF - 3
            FloatField(size=4, starting_position=680),  # Pol. Jus. REF - 4
            FloatField(size=4, starting_position=684),  # Pol. Jus. REF - 5
            FloatField(size=4, starting_position=688),  # Pol. Jus. REF - 6
            FloatField(size=4, starting_position=692),  # Canal Fuga Médio
            IntegerField(size=4, starting_position=696),  # Influência no vert
            FloatField(size=4, starting_position=700),  # Fator de Carga Máximo
            FloatField(size=4, starting_position=704),  # Fator de Carga Mínimo
            IntegerField(size=4, starting_position=708),  # Vazão Mínima Hist.
            IntegerField(size=4, starting_position=712),  # Núm. de Unid. Base
            IntegerField(size=4, starting_position=716),  # Tipo de Turbina
            IntegerField(size=4, starting_position=720),  # Repr. Conjunto
            FloatField(size=4, starting_position=724),  # TEIF
            FloatField(size=4, starting_position=728),  # IP
            IntegerField(size=4, starting_position=732),  # Tipo Perda
            LiteralField(size=12, starting_position=736),  # Data
            LiteralField(size=39, starting_position=748),  # Observação
            FloatField(size=4, starting_position=787),  # Volume de Referência
            LiteralField(size=1, starting_position=791),  # Tipo de Regular.
        ],
        storage="BINARY",
    )

    @property
    def nome(self) -> str:
        return self.data[0]

    @nome.setter
    def nome(self, v: str):
        self.data[0] = v

    @property
    def posto(self) -> int:
        return self.data[1]

    @posto.setter
    def posto(self, v: int):
        self.data[1] = v

    @property
    def subsistema(self) -> int:
        return self.data[3]

    @subsistema.setter
    def subsistema(self, v: int):
        self.data[3] = v

    @property
    def empresa(self) -> int:
        return self.data[4]

    @empresa.setter
    def empresa(self, v: int):
        self.data[4] = v

    @property
    def jusante(self) -> int:
        return self.data[5]

    @jusante.setter
    def jusante(self, v: int):
        self.data[5] = v

    @property
    def desvio(self) -> int:
        return self.data[6]

    @desvio.setter
    def desvio(self, v: int):
        self.data[6] = v

    @property
    def volume_minimo(self) -> float:
        return self.data[7]

    @volume_minimo.setter
    def volume_minimo(self, v: float):
        self.data[7] = v

    @property
    def volume_maximo(self) -> float:
        return self.data[8]

    @volume_maximo.setter
    def volume_maximo(self, v: float):
        self.data[8] = v

    @property
    def volume_vertedouro(self) -> float:
        return self.data[9]

    @volume_vertedouro.setter
    def volume_vertedouro(self, v: float):
        self.data[9] = v

    @property
    def volume_desvio(self) -> float:
        return self.data[10]

    @volume_desvio.setter
    def volume_desvio(self, v: float):
        self.data[10] = v

    @property
    def cota_minima(self) -> float:
        return self.data[11]

    @cota_minima.setter
    def cota_minima(self, v: float):
        self.data[11] = v

    @property
    def cota_maxima(self) -> float:
        return self.data[12]

    @cota_maxima.setter
    def cota_maxima(self, v: float):
        self.data[12] = v

    @property
    def polinomio_volume_cota(self) -> List[float]:
        return self.data[13:18]

    @polinomio_volume_cota.setter
    def polinomio_volume_cota(self, v: List[float]):
        self.data[13:18] = v

    @property
    def polinomio_cota_area(self) -> List[float]:
        return self.data[18:23]

    @polinomio_cota_area.setter
    def polinomio_cota_area(self, v: List[float]):
        self.data[18:23] = v

    @property
    def evaporacao(self) -> List[float]:
        return self.data[23:35]

    @evaporacao.setter
    def evaporacao(self, v: List[float]):
        self.data[23:35] = v

    @property
    def numero_conjuntos_maquinas(self) -> int:
        return self.data[35]

    @numero_conjuntos_maquinas.setter
    def numero_conjuntos_maquinas(self, v: int):
        self.data[35] = v

    @property
    def numero_maquinas_conjunto(self) -> List[int]:
        return self.data[36:41]

    @numero_maquinas_conjunto.setter
    def numero_maquinas_conjunto(self, v: List[int]):
        self.data[36:41] = v

    @property
    def potef_conjunto(self) -> List[float]:
        return self.data[41:46]

    @potef_conjunto.setter
    def potef_conjunto(self, v: List[float]):
        self.data[41:46] = v

    @property
    def hef_conjunto(self) -> List[float]:
        return self.data[47:52]

    @hef_conjunto.setter
    def hef_conjunto(self, v: List[float]):
        self.data[47:52] = v

    @property
    def qef_conjunto(self) -> List[float]:
        return self.data[52:57]

    @qef_conjunto.setter
    def qef_conjunto(self, v: List[float]):
        self.data[52:57] = v

    @property
    def produtibilidade_especifica(self) -> float:
        return self.data[57]

    @produtibilidade_especifica.setter
    def produtibilidade_especifica(self, v: float):
        self.data[57] = v

    @property
    def perdas(self) -> float:
        return self.data[58]

    @perdas.setter
    def perdas(self, v: float):
        self.data[58] = v

    @property
    def numero_polinomios_jusante(self) -> int:
        return self.data[59]

    @numero_polinomios_jusante.setter
    def numero_polinomios_jusante(self, v: int):
        self.data[59] = v

    @property
    def polinomios_jusante(self) -> List[float]:
        return self.data[60:96]

    @polinomios_jusante.setter
    def polinomios_jusante(self, v: List[float]):
        self.data[60:96] = v

    @property
    def canal_fuga_medio(self) -> float:
        return self.data[96]

    @canal_fuga_medio.setter
    def canal_fuga_medio(self, v: float):
        self.data[96] = v

    @property
    def influencia_vertimento_canal_fuga(self) -> int:
        return self.data[97]

    @influencia_vertimento_canal_fuga.setter
    def influencia_vertimento_canal_fuga(self, v: int):
        self.data[97] = v

    @property
    def fator_carga_maximo(self) -> float:
        return self.data[98]

    @fator_carga_maximo.setter
    def fator_carga_maximo(self, v: float):
        self.data[98] = v

    @property
    def fator_carga_minimo(self) -> float:
        return self.data[99]

    @fator_carga_minimo.setter
    def fator_carga_minimo(self, v: float):
        self.data[99] = v

    @property
    def vazao_minima_historica(self) -> int:
        return self.data[100]

    @vazao_minima_historica.setter
    def vazao_minima_historica(self, v: int):
        self.data[100] = v

    @property
    def numero_unidades_base(self) -> int:
        return self.data[101]

    @numero_unidades_base.setter
    def numero_unidades_base(self, v: int):
        self.data[101] = v

    @property
    def tipo_turbina(self) -> int:
        return self.data[102]

    @tipo_turbina.setter
    def tipo_turbina(self, v: int):
        self.data[102] = v

    @property
    def representacao_conjunto(self) -> int:
        return self.data[103]

    @representacao_conjunto.setter
    def representacao_conjunto(self, v: int):
        self.data[103] = v

    @property
    def teif(self) -> float:
        return self.data[104]

    @teif.setter
    def teif(self, v: float):
        self.data[104] = v

    @property
    def ip(self) -> float:
        return self.data[105]

    @ip.setter
    def ip(self, v: float):
        self.data[105] = v

    @property
    def tipo_perda(self) -> int:
        return self.data[106]

    @tipo_perda.setter
    def tipo_perda(self, v: int):
        self.data[106] = v

    @property
    def data_referencia(self) -> str:
        return self.data[107]

    @data_referencia.setter
    def data_referencia(self, v: str):
        self.data[107] = v

    @property
    def observacao(self) -> str:
        return self.data[108]

    @observacao.setter
    def observacao(self, v: str):
        self.data[108] = v

    @property
    def volume_referencia(self) -> float:
        return self.data[109]

    @volume_referencia.setter
    def volume_referencia(self, v: float):
        self.data[109] = v

    @property
    def tipo_regulacao(self) -> str:
        return self.data[110]

    @tipo_regulacao.setter
    def tipo_regulacao(self, v: str):
        self.data[110] = v



from cfinterface.components.block import Block
from cfinterface.components.line import Line

from inewave.config import MESES_DF, MAX_SERIES_SINTETICAS

from typing import IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class ValoresSerie(Block):
    """
    Bloco com a informaçao de uma tabela para o SIN, com
    entradas por série.
    """

    BEGIN_PATTERN = "     ANO: "
    END_PATTERN = "MEDIA  "
    HEADER_LINE = Line([])
    DATA_LINE = Line([])

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_ano = self.__class__.HEADER_LINE
        self.__linha = self.__class__.DATA_LINE

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ValoresSerie):
            return False
        bloco: ValoresSerie = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            cols = ["Série"] + MESES_DF + ["Média"]
            df = pd.DataFrame(tabela, columns=cols)
            df["Ano"] = self.__ano
            df = df[["Ano"] + cols]
            df = df.astype({"Série": "int64", "Ano": "int64"})
            return df

        self.__ano = self.__linha_ano.read(file.readline())[0]
        file.readline()

        # Variáveis auxiliares
        tabela = np.zeros((MAX_SERIES_SINTETICAS, len(MESES_DF) + 2))
        i = 0
        while True:
            linha = file.readline()
            if self.ends(linha):
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            tabela[i, :] = self.__linha.read(linha)
            i += 1



import inewave
from setuptools import setup, find_packages  # type: ignore

long_description = ""
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

requirements = []
with open("requirements.txt", "r") as fh:
    requirements = fh.readlines()

setup(
    name="inewave",
    version=inewave.__version__,
    author="Rogerio Alves",
    author_email="rogerioalves.ee@gmail.com",
    description="Interface para arquivos do NEWAVE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rjmalves/inewave",
    packages=find_packages(),
    package_data={"inewave": ["py.typed"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    python_requires=">=3.5",
    install_requires=requirements,
)



from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type


class GHMin(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à geração hidráulica
    mínima por usina.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = []

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="ghmin.dat") -> "GHMin":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="ghmin.dat"):
        self.write(diretorio, nome_arquivo)



from datetime import datetime
from inewave.newave.modelos.exph import BlocoUHEExph
from inewave.newave import Exph

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.exph import MockExph


def test_bloco_uhe_exph():

    m: MagicMock = mock_open(read_data="".join(MockExph))
    b = BlocoUHEExph()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 7
    assert b.data.shape[1] == 9
    assert b.data.iloc[0, 0] == 309
    assert b.data.iloc[0, 1] == "JURUENA"
    assert b.data.iloc[0, 2] == datetime(2024, 10, 1)
    assert b.data.iloc[0, 3] == 3
    assert b.data.iloc[0, 4] == 0.0
    assert b.data.iloc[1, 5] == datetime(2025, 1, 1)
    assert b.data.iloc[1, 6] == 25.0
    assert b.data.iloc[1, 7] == 1
    assert b.data.iloc[1, 8] == 1
    assert b.data.iloc[-1, -1] == 3


def test_atributos_encontrados_exph():
    m: MagicMock = mock_open(read_data="".join(MockExph))
    with patch("builtins.open", m):
        ad = Exph.le_arquivo("")
        assert ad.expansoes is not None


def test_atributos_nao_encontrados_exph():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Exph.le_arquivo("")
        assert ad.expansoes is None


def test_eq_exph():
    m: MagicMock = mock_open(read_data="".join(MockExph))
    with patch("builtins.open", m):
        ad1 = Exph.le_arquivo("")
        ad2 = Exph.le_arquivo("")
        assert ad1 == ad2


def test_neq_exph():
    m: MagicMock = mock_open(read_data="".join(MockExph))
    with patch("builtins.open", m):
        ad1 = Exph.le_arquivo("")
        ad2 = Exph.le_arquivo("")
        ad2.expansoes.iloc[0, 0] = -1
        assert ad1 != ad2


def test_leitura_escrita_exph():
    m_leitura: MagicMock = mock_open(read_data="".join(MockExph))
    with patch("builtins.open", m_leitura):
        ad1 = Exph.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        ad1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        ad2 = Exph.le_arquivo("")
        assert ad1 == ad2



from typing import Type, TypeVar, List, Optional, Union
from datetime import datetime

from cfinterface.components.register import Register
from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.eolicacadastro import (
    RegistroEolicaCadastro,
    RegistroEolicaCadastroConjuntoAerogeradores,
    RegistroEolicaCadastroAerogerador,
    RegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo,
    RegistroEolicaConjuntoAerogeradoresPotenciaEfetiva,
)


class EolicaCadastro(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao cadastro
    das usinas eólicas do sistema.
    """

    T = TypeVar("T")

    REGISTERS = [
        RegistroEolicaCadastroAerogerador,
        RegistroEolicaCadastroConjuntoAerogeradores,
        RegistroEolicaCadastro,
        RegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo,
        RegistroEolicaConjuntoAerogeradoresPotenciaEfetiva,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="eolica-cadastro.csv"
    ) -> "EolicaCadastro":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(
        self, diretorio: str, nome_arquivo="eolica-cadastro.csv"
    ):
        self.write(diretorio, nome_arquivo)

    def __registros_por_tipo(self, registro: Type[T]) -> List[T]:
        """
        Obtém os registro de um tipo, se houver algum no arquivo.

        :param registro: Um tipo de registro para ser lido
        :type registro: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int

        """
        return [b for b in self.data.of_type(registro)]

    def __obtem_registros(self, tipo: Type[T]) -> List[T]:
        return self.__registros_por_tipo(tipo)

    def __obtem_registros_com_filtros(
        self, tipo_registro: Type[T], **kwargs
    ) -> Optional[Union[T, List[T]]]:
        def __atende(r) -> bool:
            condicoes: List[bool] = []
            for k, v in kwargs.items():
                if v is not None:
                    condicoes.append(getattr(r, k) == v)
            return all(condicoes)

        regs_filtro = [
            r for r in self.__obtem_registros(tipo_registro) if __atende(r)
        ]
        if len(regs_filtro) == 0:
            return None
        elif len(regs_filtro) == 1:
            return regs_filtro[0]
        else:
            return regs_filtro

    def cria_registro(self, anterior: Register, registro: Register):
        """
        Adiciona um registro ao arquivo após um outro registro previamente
        existente.

        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.add_after(anterior, registro)

    def deleta_registro(self, registro: Register):
        """
        Remove um registro existente no arquivo.

        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.remove(registro)

    def append_registro(self, registro: Register):
        """
        Adiciona um registro ao arquivo na última posição.
        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.append(registro)

    def preppend_registro(self, registro: Register):
        """
        Adiciona um registro ao arquivo na primeira posição.
        Este método existe para retrocompatibilidade e deve ser substituído
        quando for suportado na classe :class:`RegisterFile`.
        """
        self.data.preppend(registro)

    def eolica_cadastro(
        self,
        codigo_eolica: Optional[int] = None,
        nome_eolica: Optional[str] = None,
        quantidade_conjuntos: Optional[str] = None,
    ) -> Optional[Union[RegistroEolicaCadastro, List[RegistroEolicaCadastro]]]:
        """
        Obtém um registro que cadastra uma usina eólica.

        :param codigo_eolica: código que especifica a usina
        :type codigo_eolica: int | None
        :param nome_eolica: nome da usina
        :type nome_eolica: str | None
        :param quantidade_conjuntos: quantidade de conjuntos de geradores
        :type quantidade_conjuntos: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaCadastro` |
            list[:class:`RegistroEolicaCadastro`] | None
        """
        return self.__obtem_registros_com_filtros(
            RegistroEolicaCadastro,
            codigo_eolica=codigo_eolica,
            nome_eolica=nome_eolica,
            quantidade_conjuntos=quantidade_conjuntos,
        )

    def eolica_cadastro_conjunto_aerogeradores(
        self,
        codigo_eolica: Optional[int] = None,
        indice_conjunto: Optional[int] = None,
        nome_conjunto: Optional[str] = None,
        quantidade_aerogeradores: Optional[int] = None,
    ) -> Optional[
        Union[
            RegistroEolicaCadastroConjuntoAerogeradores,
            List[RegistroEolicaCadastroConjuntoAerogeradores],
        ]
    ]:
        """
        Obtém um registro que cadastra um conjunto de aerogeradores
        de uma usina.

        :param codigo_eolica: código que especifica a usina
        :type codigo_eolica: int | None
        :param indice_conjunto: código do conjunto de geradores da usina
        :type indice_conjunto: int | None
        :param nome_conjunto: nome do conjunto
        :type nome_conjunto: str | None
        :param quantidade_aerogeradores: quantidade de geradores no conjunto
        :type quantidade_aerogeradores: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaCadastroConjuntoAerogeradores` |
            list[:class:`RegistroEolicaCadastroConjuntoAerogeradores`] | None
        """
        return self.__obtem_registros_com_filtros(
            RegistroEolicaCadastroConjuntoAerogeradores,
            codigo_eolica=codigo_eolica,
            indice_conjunto=indice_conjunto,
            nome_conjunto=nome_conjunto,
            quantidade_aerogeradores=quantidade_aerogeradores,
        )

    def eolica_cadastro_aerogerador(
        self,
        codigo_eolica: Optional[int] = None,
        indice_conjunto: Optional[int] = None,
        velocidade_cutin: Optional[float] = None,
        velocidade_nominal: Optional[float] = None,
        velocidade_cutout: Optional[float] = None,
        potencia_velocidade_cutin: Optional[float] = None,
        potencia_velocidade_nominal: Optional[float] = None,
        potencia_velocidade_cutout: Optional[float] = None,
        altura_torre: Optional[float] = None,
    ) -> Optional[
        Union[
            RegistroEolicaCadastroAerogerador,
            List[RegistroEolicaCadastroAerogerador],
        ]
    ]:
        """
        Obtém um registro que o cadastro de um conjunto de aerogeradores
        de uma usina.

        :param codigo_eolica: código que especifica a usina
        :type codigo_eolica: float | None
        :param indice_conjunto: código do conjunto de geradores da usina
        :type indice_conjunto: float | None
        :param velocidade_cutin: velocidade de cutin do gerador
        :type velocidade_cutin: float | None
        :param velocidade_nominal: velocidade nominal do gerador
        :type velocidade_nominal: float | None
        :param velocidade_cutout: velocidade de cutout do gerador
        :type velocidade_cutout: float | None
        :param potencia_velocidade_cutin: potência na velocidade de cutin
        :type potencia_velocidade_cutin: float | None
        :param potencia_velocidade_nominal: potência na velocidade nominal
        :type potencia_velocidade_nominal: float | None
        :param potencia_velocidade_cutout: potência na velocidade cutout
        :type potencia_velocidade_cutout: float | None
        :param altura_torre: altura da torre do gerador
        :type altura_torre: float | None

        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaCadastroAerogerador` |
            list[:class:`RegistroEolicaCadastroAerogerador`] | None
        """
        return self.__obtem_registros_com_filtros(
            RegistroEolicaCadastroAerogerador,
            codigo_eolica=codigo_eolica,
            indice_conjunto=indice_conjunto,
            velocidade_cutin=velocidade_cutin,
            velocidade_nominal=velocidade_nominal,
            velocidade_cutout=velocidade_cutout,
            potencia_velocidade_cutin=potencia_velocidade_cutin,
            potencia_velocidade_nominal=potencia_velocidade_nominal,
            potencia_velocidade_cutout=potencia_velocidade_cutout,
            altura_torre=altura_torre,
        )

    def eolica_conjunto_aerogeradores_quantidade_operando_periodo(
        self,
        codigo_eolica: Optional[int] = None,
        indice_conjunto: Optional[int] = None,
        periodo_inicial: Optional[datetime] = None,
        periodo_final: Optional[datetime] = None,
        numero_aerogeradores: Optional[int] = None,
    ) -> Optional[
        Union[
            RegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo,
            List[RegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo],
        ]
    ]:
        """
        Obtém um registro que o cadastro de um conjunto de aerogeradores
        de uma usina.

        :param codigo_eolica: código que especifica a usina
        :type codigo_eolica: int | None
        :param indice_conjunto: código do conjunto de geradores da usina
        :type indice_conjunto: int | None
        :param periodo_inicial: período de início de operação
        :type periodo_inicial: datetime | None
        :param periodo_final: período de fim de operação
        :type periodo_final: datetime | None
        :param numero_aerogeradores: número de geradores operando
        :type numero_aerogeradores: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo` |
            list[:class:`RegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo`] | None
        """
        return self.__obtem_registros_com_filtros(
            RegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo,
            codigo_eolica=codigo_eolica,
            indice_conjunto=indice_conjunto,
            periodo_inicial=periodo_inicial,
            periodo_final=periodo_final,
            numero_aerogeradores=numero_aerogeradores,
        )

    def eolica_conjunto_aerogeradores_potencia_efetiva_periodo(
        self,
        codigo_eolica: Optional[int] = None,
        indice_conjunto: Optional[int] = None,
        periodo_inicial: Optional[datetime] = None,
        periodo_final: Optional[datetime] = None,
        potencia_efetiva: Optional[float] = None,
    ) -> Optional[
        Union[
            RegistroEolicaConjuntoAerogeradoresPotenciaEfetiva,
            List[RegistroEolicaConjuntoAerogeradoresPotenciaEfetiva],
        ]
    ]:
        """
        Obtém um registro que contém a potência efetiva por período
        para um conjunto de aerogeradores.

        :param codigo_eolica: código que especifica a usina
        :type codigo_eolica: int | None
        :param indice_conjunto: código do conjunto de geradores da usina
        :type indice_conjunto: int | None
        :param periodo_inicial: período de início de operação
        :type periodo_inicial: datetime | None
        :param periodo_final: período de fim de operação
        :type periodo_final: datetime | None
        :param potencia_efetiva: potência efetiva do conjunto de aerogeradores
        :type potencia_efetiva: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroEolicaConjuntoAerogeradoresPotenciaEfetiva` |
            list[:class:`RegistroEolicaConjuntoAerogeradoresPotenciaEfetiva`] | None
        """
        return self.__obtem_registros_com_filtros(
            RegistroEolicaConjuntoAerogeradoresPotenciaEfetiva,
            codigo_eolica=codigo_eolica,
            indice_conjunto=indice_conjunto,
            periodo_inicial=periodo_inicial,
            periodo_final=periodo_final,
            potencia_efetiva=potencia_efetiva,
        )
